using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using MeshVR;
using UnityEngine.UI;
using Leap.Unity.Query;
using UnityEngine.EventSystems;
using UnityEngine.UI.Extensions;
using GPUTools.Common.Scripts.Tools;

namespace JayJayWon
{
    public class DCIPresetButtonPointerBehaviour : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler, IPointerClickHandler
    {

        private string infoText;
        private DAZClothingItem dci;
        public void Init(string infoText, DAZClothingItem dci)
        {
            this.infoText = infoText;
            this.dci = dci;
        }
        public void OnPointerEnter(PointerEventData eventData)
        {
            ClothItemPresetSelector.dciPresetNameJSS.val = infoText;
        }

        public void OnPointerExit(PointerEventData eventData)
        {
            ClothItemPresetSelector.dciPresetNameJSS.val = "";
        }

        void IPointerClickHandler.OnPointerClick(PointerEventData eventData)
        {
            if (eventData.clickCount == 2) dci.OpenUI();
        }

    }

    public class DCIButtonPointerBehaviour : MonoBehaviour, IPointerClickHandler
    {
        private DAZClothingItem dci;
        public void Init( DAZClothingItem dci)
        {

            this.dci = dci;
        }

        void IPointerClickHandler.OnPointerClick(PointerEventData eventData)
        {
            if (eventData.clickCount == 2) dci.OpenUI();
        }

    }


    public class ClothItemPresetSelector
    {
        private GameObject cipScrollListGO;
        private UIDynamicToggle keepOpenToggle;
        private GameObject dciPresetNameGO;

        private UIDynamicButton resetDCIToDefaultButton;

        public static JSONStorableBool keepOpenJSB = new JSONStorableBool("keepOpen", true);
        public static JSONStorableString dciPresetNameJSS = new JSONStorableString("dciPresetName", "");

        private GridsCanvas uiButtonCanvas;

        public static List<ClothingItemPreset> dciPresets;
        public static string targetPerson;
        private HashSet<UIDynamicButton> uidButtons = new HashSet<UIDynamicButton>();


        protected virtual ScrollRect scrollRect
        {
            get
            {
                if (cipScrollListGO == null) return null;
                return cipScrollListGO.GetComponent<ScrollRect>();
            }
        }
        protected Scrollbar sb { get { return scrollRect.verticalScrollbar; } }

        protected RectTransform scrollListRT { get { return cipScrollListGO.GetComponent<RectTransform>(); } }
        public ClothItemPresetSelector(GridsCanvas uiButtonCanvas)
        {
            this.uiButtonCanvas = uiButtonCanvas;
        }


        const float scrollList1Height = 500f;
        const float scrollList1Width = 400f;
        const float scrollBarWidth = 20f;
        public void CreateUI()
        {
            
            float listXoffset = -(scrollList1Width / 4f);

            cipScrollListGO = uiButtonCanvas.CreateScrollableButtonList(0f, (scrollList1Height / 2f) + 70f , scrollList1Width, scrollList1Height, new Color(0.5f, 0.5f, 0.5f, 0.5f), scrollBarWidth);


            keepOpenToggle = uiButtonCanvas.CreateUIToggleInCanvas("Keep Open", -(scrollList1Width/2f)+50f, 48f, 0f, 90f, 40f);
            keepOpenJSB.RegisterToggle(keepOpenToggle.toggle);


            dciPresetNameGO = UITools.CreateBasicTextBox(uiButtonCanvas._canvas.transform, dciPresetNameJSS, 298f, 36f, 30, -50f, 80f, anchorPreset: UITools.AnchorPresets.bottomLeft, backgoundColor: Color.gray);

            dciPresetNameJSS.text.resizeTextMinSize = 18;
            dciPresetNameJSS.text.resizeTextMaxSize = 30;
            dciPresetNameJSS.text.resizeTextForBestFit = true;
            dciPresetNameJSS.text.color = Color.white;

            dciPresetNameJSS.RegisterText(dciPresetNameJSS.text);

            resetDCIToDefaultButton = uiButtonCanvas.CreateUIButtonInCanvas("Load Defaults",-(scrollList1Width/2)+60f, scrollList1Height + 80f, 0f,120f,25f);
            resetDCIToDefaultButton.button.onClick.AddListener(delegate
            {
                if (dciPresets.Count > 0)
                {
                    Atom atom = SuperController.singleton.GetAtomByUid(targetPerson);
                    JSONStorable receiver = atom.GetStorableByID(dciPresets[0].dci.internalUid + "Preset");
                    JSONStorableUrl presetPathJSON = receiver.GetUrlJSONParam("presetBrowsePath");
                    receiver.CallAction("LoadDefaults");
                }
            });

            RefreshButtons();
        }
        private void RefreshButtons()
        {
            int buttonsPerRow = 4;
            float rowsTemp = ((float) dciPresets.Count / (float) buttonsPerRow);

            int buttonRows = (int) Math.Ceiling(rowsTemp);

            GameObject buttonListContentGO = cipScrollListGO.transform.Find("buttonListViewportGO/buttonListContentGO").gameObject;

            float buttonWidth = (scrollList1Width - (scrollBarWidth+10f)) / buttonsPerRow;

            uidButtons.Clear();

            for (int rowIndex =1; rowIndex<=buttonRows; rowIndex++)
            {
                 var cipItemContainerImageGO = UITools.CreateUIContainerImageInParentGO(buttonListContentGO.transform, scrollList1Width- scrollBarWidth, buttonWidth, new Color(0.9f, 0.9f, 0.9f,0f));
                int columnsInRow = buttonsPerRow;
                if (rowIndex == buttonRows) columnsInRow = dciPresets.Count -((buttonRows-1)*buttonsPerRow);
                for (int colIndex = 1; colIndex<=columnsInRow; colIndex++)
                {
                    int dciIndex = ((rowIndex - 1) * buttonsPerRow)+colIndex-1;
                    var presetButtonGO = UITools.CreateUIButtonInParent(cipItemContainerImageGO.transform);
                    UITools.SetSizeAndAnchors(cipItemContainerImageGO, presetButtonGO, buttonWidth, buttonWidth, "left", buttonWidth*(colIndex-1));
                    var presetButton = presetButtonGO.transform.GetComponent<UIDynamicButton>();
                    uidButtons.Add(presetButton);

                    presetButton.button.onClick.AddListener(delegate
                    {
                        dciPresets[dciIndex].LoadPreset(targetPerson);
                    });
                    dciPresets[dciIndex].LoadDCISprite(presetButton);

                    DCIPresetButtonPointerBehaviour presetInfoEnterExit = presetButton.button.gameObject.AddComponent<DCIPresetButtonPointerBehaviour>();
                    presetInfoEnterExit.Init(dciPresets[dciIndex].presetFilePathDisplayName, dciPresets[dciIndex].dci);
                }

            }
        }

        public void UpdateUI()
        {
           foreach (var button in uidButtons)
            {
                button.buttonColor = ActiveClothingEditorDisplay.grey;
            }
            keepOpenToggle.backgroundColor = ActiveClothingEditorDisplay.grey;
            resetDCIToDefaultButton.buttonColor = ActiveClothingEditorDisplay.grey;
        }

        public void DestoryUI()
        {
            if (cipScrollListGO != null) GameObject.Destroy(cipScrollListGO);
            if (keepOpenToggle != null) GameObject.Destroy(keepOpenToggle.gameObject);
            if (dciPresetNameGO != null) GameObject.Destroy(dciPresetNameGO);
            if (resetDCIToDefaultButton != null) GameObject.Destroy(resetDCIToDefaultButton.gameObject);
        }

        public void UpdateGazeTouchDistances()
        {
            GameControlUI.SetNearestGazeAndLeapButtons(keepOpenToggle);
            GameControlUI.SetNearestGazeAndLeapButtons(resetDCIToDefaultButton);

            float checkRectMinY = scrollListRT.rect.yMin;
            float checkRectMaxY = scrollListRT.rect.yMax;

            foreach (UIDynamicButton uidButton in uidButtons)
            {
                RectTransform uidButtonRT = uidButton.button.GetComponent<RectTransform>();
                Vector3 positionInWorld = uidButtonRT.parent.TransformPoint(uidButtonRT.localPosition);
                Vector3 positionInScroll = scrollListRT.InverseTransformPoint(positionInWorld);

                float childMinY = positionInScroll.y + uidButtonRT.rect.yMin;
                float childMaxY = positionInScroll.y + uidButtonRT.rect.yMax;
                if (childMaxY >= checkRectMinY && childMinY <= checkRectMaxY)
                {
                    GameControlUI.SetNearestGazeAndLeapButtons(uidButton);
                }
            }
        }
        public float GetXYDistanceToScrollBar(Vector3 worldSpacePoint, float zThreshold)
        {
            float distance = 100000f;
            RectTransform rt = sb.transform.GetComponent<RectTransform>();
            Vector3 localSpacePoint = rt.InverseTransformPoint(worldSpacePoint);

            if (localSpacePoint.z < zThreshold) distance = GeometryUtils.DistancePointToRectangle(new Vector2(localSpacePoint.x, localSpacePoint.y), rt.rect);

            return (distance);
        }

        public float GetScrollBarValue()
        {
            return (sb.value);
        }

        public void OffsetScrollBarValueByYMovement(float initialSBValue, Vector3 initialWorldSpacePoint, Vector3 currentWorldSpacePoint)
        {
            RectTransform rt = sb.transform.GetComponent<RectTransform>();
            Vector3 localInitialSpacePoint = rt.InverseTransformPoint(initialWorldSpacePoint);
            Vector3 localCurrentSpacePoint = rt.InverseTransformPoint(currentWorldSpacePoint);

            sb.value = Mathf.Clamp(initialSBValue + ((localCurrentSpacePoint.y - localInitialSpacePoint.y) / rt.rect.height), 0f, 1f);
        }

    }

    public class ClothingItemPreset
    {
        public DAZClothingItem dci { get; private set; }
        public string presetFilePath { get; private set; }
        private string textureFilePath { get { return presetFilePath.Substring(0, presetFilePath.Length - 4) + ".jpg"; } }

        private string presetFileDisplayName { get {
                string prefix = dci.dynamicRuntimeLoadPath.Substring(0, dci.dynamicRuntimeLoadPath.Length - 4);
                
                int startIndex = prefix.Length+1;
                return presetFilePath.Substring(startIndex, presetFilePath.Length - startIndex - 4);

            } }

        public string presetFilePathDisplayName { get
            {
                string fileName = FileManagerSecure.GetFileName(presetFilePath);
                if (dci.packageUid != null && dci.packageUid != "")  return dci.packageUid + ": " + fileName;
                return fileName;
            } }

        private bool textureLoading = false;
        private UIDynamicButton uidButtonImageLoad;

        private static Dictionary<string, Texture2D> dciPresetTexturesCache = new Dictionary<string, Texture2D>();
        public static void ClearCachedPresetTextures()
        {
            foreach (var kvp in dciPresetTexturesCache)
            {
                GameObject.Destroy(kvp.Value);
            }
            dciPresetTexturesCache.Clear();
        }

        public ClothingItemPreset(DAZClothingItem dci, string presetFilePath)
        {
            this.dci = dci;
            this.presetFilePath = presetFilePath;
        }

        public void LoadPreset(string personName)
        {
            Atom atom = SuperController.singleton.GetAtomByUid(personName);
            JSONStorable receiver = atom.GetStorableByID(dci.internalUid + "Preset");
            JSONStorableUrl presetPathJSON = receiver.GetUrlJSONParam("presetBrowsePath");
            presetPathJSON.val = presetFilePath;
            receiver.CallAction("LoadPreset");

            if (!ClothItemPresetSelector.keepOpenJSB.val)
            {
                GameControlUI.gameControlDisplayMode = GameControlDisplayModes.ace;
                GameControlUI.RefreshWristUIButtonGrid();
            }
        }

        public void LoadDCISprite(UIDynamicButton uidButton)
        {
            if (dciPresetTexturesCache.ContainsKey(textureFilePath))
            {

                SetDCITexture(dciPresetTexturesCache[textureFilePath], uidButton);
                uidButton.label = "";
            }
            else
            {
                if (FileManagerSecure.FileExists(textureFilePath) && !textureLoading)
                {
                    uidButtonImageLoad = uidButton;
                    textureLoading = true;
                    ImageUtils.QueueLoadTexture(textureFilePath, QueuedImageLoadCallback, true);
                    uidButton.label = "";
                }
                else
                {
                    if (!FileManagerSecure.FileExists(textureFilePath))
                    {
                        uidButton.label = presetFileDisplayName;
                        uidButton.buttonText.resizeTextMinSize = 12;
                        uidButton.buttonText.resizeTextMaxSize = 16;
                        uidButton.buttonText.resizeTextForBestFit = true;

                    }
                }
            }        
        }

        private void QueuedImageLoadCallback(ImageLoaderThreaded.QueuedImage qi)
        {
            Texture2D tex = qi.tex;
            if (tex != null)
            {
                dciPresetTexturesCache.Add(textureFilePath, tex);
                SetDCITexture(tex, uidButtonImageLoad);
            }
            textureLoading = false;
        }

        private void SetDCITexture(Texture2D tex, UIDynamicButton targetButton)
        {
            Sprite mySprite = Sprite.Create(tex, new Rect(0.0f, 0.0f, tex.width, tex.height), new Vector2(0.5f, 0.5f), 100.0f);
            targetButton.button.image.sprite = mySprite;
        }
    }
    public class ACEPlusDisplay: ActiveClothingEditorDisplay
    {
        public int aceMode = ACEMode.original;

        protected UIDynamicButton aceSelectGroupToggleAllButton;
        protected UIDynamicButton aceSelectGroupRemoveToggleAllButton;
        protected UIDynamicButton aceSelectGroupDressToggleAllButton;
        protected UIDynamicButton aceSelectGroupResetSimAllButton;
        protected GameObject aceSelectStatusGO;

        protected UIDynamicPopup aceTargetAtomPopup;
        protected UIDynamicButton aceClearUnwornClothingButton;
        protected GameObject aceTargetAtomTextGO;


        public static Dictionary<DAZClothingItem, List<ClothingItemPreset>> clothingPresets = new Dictionary<DAZClothingItem, List<ClothingItemPreset>>();

        protected static Dictionary<string, float> atomACEScrollBarPosition = new Dictionary<string, float>();

        protected JSONStorableString aceSelectStatusJSS;
        protected JSONStorableString aceTargetAtomTextJSS;

        protected bool scrollbarListenerCreated = false;

        public JSONStorableStringChooser personAtomNamesJSSC;

        public bool isGazeSelectedMode = false;
        public int gazeSelectedTargetType = LastViewedTargetType.lastViewedFemale;

        public ACEPlusDisplay(GridsCanvas uiButtonCanvas):base(uiButtonCanvas)
        {
            aceSelectStatusJSS = new JSONStorableString("aceSelectStatus", "");
            personAtomNamesJSSC = new JSONStorableStringChooser("personAtomNames", null, "","Target Person");
            aceTargetAtomTextJSS = new JSONStorableString("aceTargetAtomText", "");
        }

        public override void AtomRemovedUpdate(string atomName)
        {
            base.AtomRemovedUpdate(atomName);
            atomACEScrollBarPosition.Remove(atomName);
            if (aceMode == ACEMode.enhanced)
            {
                RefreshPersonAtomNames();
                if (personAtomNamesJSSC.val == atomName) personAtomNamesJSSC.val = "";
            }
            if (GameControlUI.gameControlDisplayMode==GameControlDisplayModes.ace) RefreshACE();
        }

        public override void UpdateAtomNames(string oldName, string newName)
        {
            base.UpdateAtomNames(oldName, newName);
            if (atomACEScrollBarPosition.ContainsKey(oldName))
            {
                float tempScrollBarPos = atomACEScrollBarPosition[oldName];
                atomACEScrollBarPosition.Remove(oldName);
                atomACEScrollBarPosition.Add(newName, tempScrollBarPos);
            }

            if (aceMode == ACEMode.enhanced)
            {
                if (personAtomNamesJSSC.val == oldName) personAtomNamesJSSC.val = newName;
                RefreshPersonAtomNames();
            }
        }


        public void RefreshPersonAtomNames()
        {
            var personAtoms = AtomUtils.GetPersonAtoms();
            List<string> personAtomNames = new List<string>();
            List<string> personAtomDisplayNames = new List<string>();

            foreach (Atom atom in personAtoms)
            {
                personAtomNames.Add(atom.name);
                if (AtomUtils.IsMale(atom)) personAtomDisplayNames.Add(atom.name + " [M]");
                else personAtomDisplayNames.Add(atom.name + " [F]");
            }
            personAtomNamesJSSC.choices = personAtomNames;
            personAtomNamesJSSC.displayChoices = personAtomDisplayNames;

            if (personAtomNames.Count > 0)
            {
                if (!personAtomNames.Contains(personAtomNamesJSSC.val)) personAtomNamesJSSC.val = personAtomNames[0];
            }
            else personAtomNamesJSSC.val = "";
        }

        public new void CreateUI()
        {
            if (aceMode == ACEMode.original) base.CreateUI();
            else
            {
                clothingPresets.Clear();
                float scrollList1Height = 500f;
                float scrollList1Width = 400f;
                float listXoffset = -(scrollList1Width / 4f);
                float selectGroupHeight = 25f;

                aceScrollListGO = _uiButtonCanvas.CreateScrollableButtonList(0f, (scrollList1Height / 2f) + _verticalStartOffset + selectGroupHeight, scrollList1Width, scrollList1Height, new Color(0.5f, 0.5f, 0.5f, 0.5f), 20f);

                GameObject buttonListContentGO = aceScrollListGO.transform.Find("buttonListViewportGO/buttonListContentGO").gameObject;
                VerticalLayoutGroup vlg = buttonListContentGO.GetComponent<VerticalLayoutGroup>();
                vlg.spacing = 3f;

                aceSelectGroupToggleAllButton = _uiButtonCanvas.CreateUIButtonInCanvas("Select All", -scrollList1Width / 2 + 40f, _verticalStartOffset + 12.5f, 0f, 80f, 25f,14);
                aceSelectGroupToggleAllButton.button.image.sprite = null;
                aceSelectGroupToggleAllButton.button.onClick.AddListener(delegate
                {
                    try { SelectAllToggled(); }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                });

                aceSelectGroupResetSimAllButton = _uiButtonCanvas.CreateUIButtonInCanvas("Reset\nSim", scrollList1Width / 2f - 25f, _verticalStartOffset + 12.5f, 0f, 50f, 25f);
                aceSelectGroupResetSimAllButton.button.image.sprite = null;
                aceSelectGroupResetSimAllButton.buttonText.fontSize = 11;
                aceSelectGroupResetSimAllButton.buttonColor = Color.red;
                aceSelectGroupResetSimAllButton.textColor = Color.white;
                aceSelectGroupResetSimAllButton.button.onClick.AddListener(delegate
                {
                    try { ResetSimAllSelected(); }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                });

                aceSelectGroupDressToggleAllButton = _uiButtonCanvas.CreateUIButtonInCanvas("Undress\nSelected", scrollList1Width / 2f - 25f-50f, _verticalStartOffset + 12.5f, 0f, 50f, 25f);
                aceSelectGroupDressToggleAllButton.button.image.sprite = null;
                aceSelectGroupDressToggleAllButton.buttonText.fontSize = 11;
                aceSelectGroupDressToggleAllButton.button.onClick.AddListener(delegate
                {
                    try { ToggleUndressAllSelected(); }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                });

                aceSelectGroupRemoveToggleAllButton = _uiButtonCanvas.CreateUIButtonInCanvas("Remove\nSelected", scrollList1Width / 2f - 25f-100f, _verticalStartOffset + 12.5f, 0f, 50f, 25f);
                aceSelectGroupRemoveToggleAllButton.button.image.sprite = null;
                aceSelectGroupRemoveToggleAllButton.buttonText.fontSize = 11;
                aceSelectGroupRemoveToggleAllButton.button.onClick.AddListener(delegate
                {
                    try { ToggleRemoveAllSelected(); }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                });

                aceSelectStatusGO = UITools.CreateBasicTextBox(_uiButtonCanvas._canvas.transform, aceSelectStatusJSS, 170f, 25f, 30, -scrollList1Width / 2f + 130f, _verticalStartOffset + 50f, anchorPreset: UITools.AnchorPresets.bottomLeft, backgoundColor: Color.cyan);

                if (!isGazeSelectedMode)
                {
                    aceTargetAtomPopup = UITools.CreateScrollablePopup(_uiButtonCanvas._canvas.transform, personAtomNamesJSSC, 280f, 30f, "Target Person", 28, -scrollList1Width / 2f + 140f, scrollList1Height + 20f, 300f, false, true, popupRowHeight:25f, popupFontSize:28);
                    RefreshPersonAtomNames();
                    aceTargetAtomPopup.popup.onOpenPopupHandlers += () =>
                    {
                        RefreshPersonAtomNames();
                    };
                }
                else
                {
                    aceTargetAtomTextGO = UITools.CreateBasicTextBox(_uiButtonCanvas._canvas.transform, aceTargetAtomTextJSS, 170f, 25f, 30, -scrollList1Width / 2f+ 50f, _verticalStartOffset + 77f + scrollList1Height, anchorPreset: UITools.AnchorPresets.bottomLeft,backgoundColor:Color.gray);
                }
                


                aceClearUnwornClothingButton = _uiButtonCanvas.CreateUIButtonInCanvas("Clear Unworn Items", scrollList1Width / 2f - 50f, scrollList1Height + 42f+_verticalStartOffset, 0f, 100f, 28f);
                aceClearUnwornClothingButton.buttonText.fontSize = 12;
                aceClearUnwornClothingButton.button.image.sprite = null;



                aceClearUnwornClothingButton.button.onClick.AddListener(delegate
                {
                    try { RefreshClothing(_currentACEAtomName); }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                });

                aceScrollBarPositionRefreshRequired = true;
            }
        }

        protected void SelectAllToggled()
        {
            ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
            currentACL.ToggleAllSelectedDCIStates();
        }
        protected void ResetSimAllSelected()
        {
            ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
            currentACL.ResetSimAllSelected();
        }

        protected void ToggleUndressAllSelected()
        {
            ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
            currentACL.ToggleUndressAllSelected();
        }

        protected void ToggleRemoveAllSelected()
        {
            ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
            Atom targetAtom = SuperController.singleton.GetAtomByUid(_currentACEAtomName);
            if (targetAtom!=null) currentACL.ToggleRemoveAllSelected(targetAtom);
        }

        public override void DestroyUI()
        {
            if (scrollRect!=null) scrollRect.verticalScrollbar.onValueChanged.RemoveAllListeners();
            base.DestroyUI();
            if (aceSelectGroupToggleAllButton!=null) GameObject.Destroy(aceSelectGroupToggleAllButton.gameObject);

            if (aceSelectGroupDressToggleAllButton != null) GameObject.Destroy(aceSelectGroupDressToggleAllButton.gameObject);
            if (aceSelectGroupResetSimAllButton != null) GameObject.Destroy(aceSelectGroupResetSimAllButton.gameObject);
            if (aceSelectGroupRemoveToggleAllButton != null) GameObject.Destroy(aceSelectGroupRemoveToggleAllButton.gameObject);
            if (aceSelectStatusGO != null) GameObject.Destroy(aceSelectStatusGO);
            if (aceTargetAtomTextGO != null) GameObject.Destroy(aceTargetAtomTextGO);

            if (aceTargetAtomPopup != null) GameObject.Destroy(aceTargetAtomPopup.gameObject);

            if (aceClearUnwornClothingButton != null) GameObject.Destroy(aceClearUnwornClothingButton.gameObject);

            scrollbarListenerCreated = false;
        }

        public override void UpdateACEGazeTouchDistances()
        {
            if (aceMode == ACEMode.original) base.UpdateACEGazeTouchDistances();
            else
            {
                GameControlUI.SetNearestGazeAndLeapButtons(aceSelectGroupToggleAllButton);

                GameControlUI.SetNearestGazeAndLeapButtons(aceSelectGroupDressToggleAllButton);
                GameControlUI.SetNearestGazeAndLeapButtons(aceSelectGroupResetSimAllButton);
                GameControlUI.SetNearestGazeAndLeapButtons(aceSelectGroupRemoveToggleAllButton);
                GameControlUI.SetNearestGazeAndLeapButtons(aceClearUnwornClothingButton);

                float checkRectMinY = scrollListRT.rect.yMin;
                float checkRectMaxY = scrollListRT.rect.yMax;

                foreach (ACEIDPlus uidACEitem in _aceScrollListItems)
                {
                    RectTransform selectToggleRT = uidACEitem.aceItemContainerImageGO.GetComponent<RectTransform>();
                    Vector3 positionInWorld = selectToggleRT.parent.TransformPoint(selectToggleRT.localPosition);
                    Vector3 positionInScroll = scrollListRT.InverseTransformPoint(positionInWorld);

                    float childMinY = positionInScroll.y + selectToggleRT.rect.yMin;
                    float childMaxY = positionInScroll.y + selectToggleRT.rect.yMax;

                    if (childMaxY >= checkRectMinY && childMinY <= checkRectMaxY)
                    {
                        GameControlUI.SetNearestGazeAndLeapButtons(uidACEitem.aceItemImageButton);
                        GameControlUI.SetNearestGazeAndLeapButtons(uidACEitem.aceItemNameButton);
                        GameControlUI.SetNearestGazeAndLeapButtons(uidACEitem.aceItemPresetsButton);
                         GameControlUI.SetNearestGazeAndLeapButtons(uidACEitem.aceItemResetSimButton);
                        GameControlUI.SetNearestGazeAndLeapButtons(uidACEitem.aceItemRemoveButton);
                        GameControlUI.SetNearestGazeAndLeapButtons(uidACEitem.aceItemUndressButton);

                    }
                }

            }
        }



        private IEnumerator SetScollBarPosition(float newScrollBarPosition)
        {
            yield return new WaitForEndOfFrame();
            if (scrollRect != null) scrollRect.verticalNormalizedPosition = newScrollBarPosition;
        }

        public void UpdateACE(string atomName, ActiveClothingLists currentActiveClothing)
        {
            if (aceMode == ACEMode.original && _currentACEAtomName != "") StoreScrollBarPosition(_currentACEAtomName);
            if (atomName != "" && !SuperController.singleton.isLoading)
            {
                if (!_atomACEClothingLists.ContainsKey(atomName)) _atomACEClothingLists.Add(atomName, new ActiveClothingLists());
                

                ActiveClothingLists currentACEClothingLists = _atomACEClothingLists[atomName];
                List<DAZClothingItem> currentACEClothingItems = currentACEClothingLists._activeClothingDCIs;
                List<DAZClothingItem> currentActiveClothingItems = currentActiveClothing._activeClothingDCIs;
                List<DAZClothingItem> missingActiveClothing = new List<DAZClothingItem>();

                if (currentACEClothingItems.Count > 0)
                {
                    var temp = currentActiveClothingItems.Except(currentACEClothingItems);
                    if (temp.Count<DAZClothingItem>() > 0)
                    {
                        foreach (DAZClothingItem item in temp) missingActiveClothing.Add(item);
                    }
                }
                else missingActiveClothing = currentActiveClothingItems;

                if (missingActiveClothing.Count > 0)
                {
                    foreach (DAZClothingItem clothingItem in missingActiveClothing)
                    {
                        currentACEClothingItems.Add(clothingItem);

                        if (currentActiveClothing._activeClothingTags.ContainsKey(clothingItem)) currentACEClothingLists._activeClothingTags.Add(clothingItem, currentActiveClothing._activeClothingTags[clothingItem]);

                        currentACEClothingLists._activeClothingDisplayNames.Add(clothingItem, currentActiveClothing._activeClothingDisplayNames[clothingItem]);

                        if (currentActiveClothing._activeClothingCSC.ContainsKey(clothingItem)) _atomACEClothingLists[atomName]._activeClothingCSC.Add(clothingItem, currentActiveClothing._activeClothingCSC[clothingItem]);

                        currentACEClothingLists._activeClothingSelected.Add(clothingItem, false);
                    }
                }
                if (atomName != _currentACEAtomName) aceScrollBarPositionRefreshRequired = true;

                if (missingActiveClothing.Count > 0 || atomName != _currentACEAtomName)
                {
                    _currentACEAtomName = atomName;
                    SetUndressSelectedButtonState();
                    SetRemoveSelectedButtonState();
                }

            }
            else _currentACEAtomName = "";

            RefreshACE();
            if (aceMode == ACEMode.original) RestoreScrollBarPosition(_currentACEAtomName);
        }


        protected void RefreshClothingPresets(DAZClothingItem dci)
        {
            ClothingItemPreset.ClearCachedPresetTextures();
            if (clothingPresets.ContainsKey(dci)) clothingPresets[dci].Clear();
            else clothingPresets.Add(dci, new List<ClothingItemPreset>());

            string folderName = FileManagerSecure.GetDirectoryName(dci.uid);

            if (FileManagerSecure.DirectoryExists(folderName)) AddClothingPresets(dci, clothingPresets[dci],folderName);
            
            if (dci.packageUid!=null && dci.packageUid!="")
            {
                int index = folderName.IndexOf(".var:");
                string localFolderName = folderName.Substring(index+6);
                if (FileManagerSecure.DirectoryExists(localFolderName)) AddClothingPresets(dci, clothingPresets[dci], localFolderName);
            }
            
        }

        protected void AddClothingPresets(DAZClothingItem dci, List<ClothingItemPreset> clothingIteamPresets, string folderPath)
        {
            var files = FileManagerSecure.GetFiles(folderPath, "*.vap");
            foreach (var file in files)
            {
                string normalPath = FileManagerSecure.NormalizePath(file);
                string prefix =  dci.dynamicRuntimeLoadPath.Substring(0, dci.dynamicRuntimeLoadPath.Length - 4);

                if (normalPath.StartsWith(prefix + "_"))
                {
                    bool isHidden = false;
                    string hideFlagFilePath="";
                    if (FileManagerSecure.IsFileInPackage(file))
                    {
                        string varFolderPath = FileManagerSecure.GetDirectoryName(file.Substring(file.IndexOf(':')+1));
                        string varFileName = FileManagerSecure.GetFileName(file) + ".hide";
                        hideFlagFilePath = "AddonPackagesFilePrefs\\" + dci.packageUid +  varFolderPath+"\\"+ varFileName;
                        if (FileManagerSecure.FileExists(hideFlagFilePath)) isHidden = true;
                    }
                    else if (FileManagerSecure.FileExists(file+".hide")) isHidden = true;

                    if (!isHidden) clothingIteamPresets.Add(new ClothingItemPreset(dci, normalPath));
                }
            }
        }

        protected override void RefreshACE()
        {
            if (aceMode == ACEMode.original) base.RefreshACE();
            else
            {
                if (personAtomNamesJSSC.val == "") RefreshPersonAtomNames();
                if (_atomACEClothingLists.ContainsKey(_currentACEAtomName))
                {
                    SetClothingItemCount(_atomACEClothingLists[_currentACEAtomName]._activeClothingDCIs.Count);
                    ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
                    int index = 0;
                    foreach (ACEIDPlus uidACEitem in _aceScrollListItems)
                    {
                        DAZClothingItem dci = currentACL._activeClothingDCIs[index];
                        if (!clothingPresets.ContainsKey(dci)) RefreshClothingPresets(dci); ;
                        uidACEitem.RefreshACEItem(currentACL, index);
                        
                        index++;
                    }

                    if (!scrollbarListenerCreated)
                    {
                        scrollRect.verticalScrollbar.onValueChanged.AddListener(delegate
                        {
                            if (!atomACEScrollBarPosition.ContainsKey(_currentACEAtomName)) atomACEScrollBarPosition.Add(_currentACEAtomName, 1f);
                            if (scrollRect.verticalScrollbar.size > 0.999f) atomACEScrollBarPosition[_currentACEAtomName] = 1f;
                            else atomACEScrollBarPosition[_currentACEAtomName] = scrollRect.verticalNormalizedPosition;

                        });
                        scrollbarListenerCreated = true;
                    }

                    if (!atomACEScrollBarPosition.ContainsKey(_currentACEAtomName)) atomACEScrollBarPosition.Add(_currentACEAtomName, 1f);

                    if (aceScrollBarPositionRefreshRequired)
                    {
                        UIAGlobals.mvrScript.StartCoroutine(SetScollBarPosition(atomACEScrollBarPosition[_currentACEAtomName]));
                        aceScrollBarPositionRefreshRequired = false;
                    }
                    RefreshACEGroupGeneral(currentACL);

                }
                else
                {
                    SetClothingItemCount(0);
                    RefreshACEGroupGeneral(new ActiveClothingLists());
                }
                
            }
                
        }

        protected void RefreshACEGroupGeneral(ActiveClothingLists currentACL)
        {

            if (aceMode == ACEMode.original) base.RefreshACEGroup(currentACL);
            else
            {
                if (currentACL._activeClothingDCIs.Count == 0)
                {
                    aceSelectGroupToggleAllButton.gameObject.SetActive(false);
                    aceSelectGroupDressToggleAllButton.gameObject.SetActive(false);
                    aceSelectGroupResetSimAllButton.gameObject.SetActive(false);
                    aceSelectGroupRemoveToggleAllButton.gameObject.SetActive(false);
                    aceSelectStatusGO.SetActive(false);
                    aceClearUnwornClothingButton.gameObject.SetActive(false);
                    if (aceTargetAtomTextGO!=null) aceTargetAtomTextGO.SetActive(false);
                }
                else
                {
                    if (aceTargetAtomTextGO != null) aceTargetAtomTextGO.SetActive(true);
                    aceTargetAtomTextJSS.val = _currentACEAtomName;

                    aceSelectGroupToggleAllButton.gameObject.SetActive(true);
                    aceSelectGroupToggleAllButton.buttonColor = grey;

                    if (currentACL.isAllDCIWorn) aceClearUnwornClothingButton.gameObject.SetActive(false);
                    else aceClearUnwornClothingButton.gameObject.SetActive(true);
                    if (currentACL.isAllDCISelected) aceSelectGroupToggleAllButton.label = "Unselect All";
                    else aceSelectGroupToggleAllButton.label = "Select All";
                    if (currentACL.selectedCount == 0)
                    {
                        aceSelectGroupDressToggleAllButton.gameObject.SetActive(false);
                        aceSelectGroupResetSimAllButton.gameObject.SetActive(false);
                        aceSelectGroupRemoveToggleAllButton.gameObject.SetActive(false);
                        aceSelectStatusGO.SetActive(false);
                    }
                    else
                    {
                        aceSelectGroupRemoveToggleAllButton.gameObject.SetActive(true);
                        aceSelectStatusGO.SetActive(true);

                        aceSelectStatusJSS.val = currentACL.selectedCount.ToString() + " Selected Items";
                        if (currentACL.isAllSelectedDCIWorn)
                        {
                            aceSelectGroupRemoveToggleAllButton.label = "Remove\nSelected";
                            aceSelectGroupRemoveToggleAllButton.buttonColor = removeOn;
                        }
                        else
                        {
                            aceSelectGroupRemoveToggleAllButton.label = "Wear\nSelected";
                            aceSelectGroupRemoveToggleAllButton.buttonColor = removeOff;
                        }

                        if (currentACL.isAnySelectDCISimEnabled)
                        {
                            aceSelectGroupDressToggleAllButton.gameObject.SetActive(true);
                            if (currentACL.isAnySelectDCINotUndressing)
                            {
                                aceSelectGroupDressToggleAllButton.label = "Undress\nSelected";
                                aceSelectGroupDressToggleAllButton.buttonColor = undressOff;
                            }
                            else
                            {
                                aceSelectGroupDressToggleAllButton.label = "Stop Undress";
                                aceSelectGroupDressToggleAllButton.buttonColor = undressOn;
                            }
                            aceSelectGroupResetSimAllButton.gameObject.SetActive(true);
                            aceSelectGroupResetSimAllButton.buttonColor = Color.red;
                            aceSelectGroupResetSimAllButton.textColor = Color.white;
                        }
                        else
                        {
                            aceSelectGroupDressToggleAllButton.gameObject.SetActive(false);
                            aceSelectGroupResetSimAllButton.gameObject.SetActive(false);
                        }
                    }
                }
                
            }
        }

        protected override void AddACEID(int listIndex)
        {

            if (aceMode == ACEMode.original) base.AddACEID(listIndex);
            else
            {
                
                ACEIDPlus aceItem = new ACEIDPlus(_uiButtonCanvas, aceScrollListGO, new Color(0.5f, 0.5f, 0f, 0f));


                aceItem.aceItemImageButton.button.onClick.AddListener(delegate
                {
                    try { SelectToggled(listIndex); }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                });

                aceItem.aceItemNameButton.button.onClick.AddListener(delegate
                {
                    try { SelectToggled(listIndex); }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                });


                aceItem.aceItemRemoveButton.button.onClick.AddListener(delegate
                {
                    try { RemoveButtonActivated(listIndex); }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                });

                aceItem.aceItemUndressButton.button.onClick.AddListener(delegate
                {
                    try { UndressButtonActivated(listIndex, false); }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                });

                aceItem.aceItemResetSimButton.button.onClick.AddListener(delegate
                {
                    try { ResetSimButtonActivated(listIndex); }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                });
                aceItem.aceItemPresetsButton.button.onClick.AddListener(delegate
                {
                    try { CIPSelectActivated(listIndex); }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                });

                _aceScrollListItems.Add(aceItem);
            }
        }

        protected void CIPSelectActivated(int index)
        {
            DAZClothingItem dci = _atomACEClothingLists[_currentACEAtomName]._activeClothingDCIs[index];
            RefreshClothingPresets(dci);

            if (clothingPresets.ContainsKey(dci))
            {
                var dciClothPresets = clothingPresets[dci];
                if (dciClothPresets.Count > 0)
                {
                    ClothItemPresetSelector.dciPresets = dciClothPresets;
                    ClothItemPresetSelector.targetPerson = _currentACEAtomName;

                    GameControlUI.gameControlDisplayMode = GameControlDisplayModes.clothItemPresetSelect;
                    
                }
            }
            GameControlUI.RefreshWristUIButtonGrid();


        }
        protected void ResetSimButtonActivated(int index)
        {
            DAZClothingItem dci = _atomACEClothingLists[_currentACEAtomName]._activeClothingDCIs[index];
            ClothSimControl csc = _atomACEClothingLists[_currentACEAtomName]._activeClothingCSC[dci];

            Atom atom = SuperController.singleton.GetAtomByUid(_currentACEAtomName);
            if (atom != null)
            {
                JSONStorable receiver = atom.GetStorableByID(csc.storeId);
                if (receiver != null) csc.Reset();
            }
        }

    }
    public class ActiveClothingEditorDisplay
    {
        protected GridsCanvas _uiButtonCanvas;
        protected GameObject aceScrollListGO;
        protected List<ActiveClothingEditorItemDisplay> _aceScrollListItems;
        protected ActiveClothingEditorItemDisplay _aceSelectedItemsGroup;
        protected UIDynamicButton _refreshButton;
        protected UIDynamicTextField _targetAtomTextField;


        protected string _currentACEAtomName = "";
        protected Dictionary<string, ActiveClothingLists> _atomACEClothingLists;
        protected Dictionary<string, Vector2> _atomACEScrollBarRectPosition;

        protected bool _removeSelectedGroupState;
        protected bool _undressSelectedGroupState;
        protected bool _suppressToggleGroupSelectOn = false;
        protected float _verticalStartOffset = 30f;

        protected bool aceScrollBarPositionRefreshRequired = false;

        protected virtual ScrollRect scrollRect { get {
                if (aceScrollListGO == null) return null;
                return aceScrollListGO.GetComponent<ScrollRect>(); } }
        protected Scrollbar sb { get { return scrollRect.verticalScrollbar; } }
        protected RectTransform scrollListRT { get { return aceScrollListGO.GetComponent<RectTransform>(); } }

        protected ScrollRect scrollListSR
        {
            get
            {
                return aceScrollListGO.GetComponent<ScrollRect>();
            }
        }

        public ActiveClothingEditorDisplay(GridsCanvas uiButtonCanvas)
        {
            _uiButtonCanvas = uiButtonCanvas;
            _aceScrollListItems = new List<ActiveClothingEditorItemDisplay>();
            _atomACEClothingLists = new Dictionary<string, ActiveClothingLists>();
            _atomACEScrollBarRectPosition = new Dictionary<string, Vector2>();
        }

        public float GetXYDistanceToScrollBar(Vector3 worldSpacePoint, float zThreshold)
        {
            float distance = 100000f;
            RectTransform rt = sb.transform.GetComponent<RectTransform>();
            Vector3 localSpacePoint = rt.InverseTransformPoint(worldSpacePoint);

            if (localSpacePoint.z < zThreshold) distance = GeometryUtils.DistancePointToRectangle(new Vector2(localSpacePoint.x, localSpacePoint.y), rt.rect);

            return (distance);
        }

        public float GetScrollBarValue()
        {
            return (sb.value);
        }
        public void OffsetScrollBarValueByYMovement(float initialSBValue, Vector3 initialWorldSpacePoint, Vector3 currentWorldSpacePoint)
        {
            RectTransform rt = sb.transform.GetComponent<RectTransform>();
            Vector3 localInitialSpacePoint = rt.InverseTransformPoint(initialWorldSpacePoint);
            Vector3 localCurrentSpacePoint = rt.InverseTransformPoint(currentWorldSpacePoint);

            sb.value = Mathf.Clamp(initialSBValue + ((localCurrentSpacePoint.y - localInitialSpacePoint.y) / rt.rect.height), 0f, 1f);
        }

        public virtual void UpdateACEGazeTouchDistances()
        {
            GameControlUI.SetNearestGazeAndLeapButtons(_refreshButton);
            GameControlUI.SetNearestGazeAndLeapButtons(_aceSelectedItemsGroup.aceItemRemoveButton);
            GameControlUI.SetNearestGazeAndLeapButtons(_aceSelectedItemsGroup.aceItemUndressButton);
            GameControlUI.SetNearestGazeAndLeapButtons(_aceSelectedItemsGroup._aceItemSelectToggle);

            float checkRectMinY = scrollListRT.rect.yMin;
            float checkRectMaxY = scrollListRT.rect.yMax;

            foreach (ActiveClothingEditorItemDisplay uidACEitem in _aceScrollListItems)
            {
                RectTransform selectToggleRT = uidACEitem.aceItemContainerImageGO.GetComponent<RectTransform>();
                Vector3 positionInWorld = selectToggleRT.parent.TransformPoint(selectToggleRT.localPosition);
                Vector3 positionInScroll = scrollListRT.InverseTransformPoint(positionInWorld);

                float childMinY = positionInScroll.y + selectToggleRT.rect.yMin;
                float childMaxY = positionInScroll.y + selectToggleRT.rect.yMax;

                if (childMaxY >= checkRectMinY && childMinY <= checkRectMaxY)
                {
                    GameControlUI.SetNearestGazeAndLeapButtons(uidACEitem.aceItemRemoveButton);
                    GameControlUI.SetNearestGazeAndLeapButtons(uidACEitem.aceItemUndressButton);
                    GameControlUI.SetNearestGazeAndLeapButtons(uidACEitem._aceItemSelectToggle);

                }
            }

        }
        public virtual void AtomRemovedUpdate(string atomName)
        {
            _atomACEClothingLists.Remove(atomName);
            _atomACEScrollBarRectPosition.Remove(atomName);

            if (_currentACEAtomName == atomName) _currentACEAtomName = "";
        }

        public virtual void UpdateAtomNames(string oldName, string newName)
        {
            if (_currentACEAtomName == oldName) _currentACEAtomName = newName;

            if (_atomACEClothingLists.ContainsKey(oldName))
            {
                ActiveClothingLists tempACL = _atomACEClothingLists[oldName];
                _atomACEClothingLists.Remove(oldName);
                _atomACEClothingLists.Add(newName, tempACL);
            }
            if (_atomACEScrollBarRectPosition.ContainsKey(oldName))
            {
                Vector2 tempScrollBarPos = _atomACEScrollBarRectPosition[oldName];
                _atomACEScrollBarRectPosition.Remove(oldName);
                _atomACEScrollBarRectPosition.Add(newName, tempScrollBarPos);
            }
        }
        protected void StoreScrollBarPosition(string atomName)
        {
//            Canvas.ForceUpdateCanvases();
            if (atomName != "")
            {
                if (_atomACEScrollBarRectPosition.ContainsKey(atomName)) _atomACEScrollBarRectPosition[atomName] = scrollListSR.content.localPosition;
                else _atomACEScrollBarRectPosition.Add(atomName, scrollListSR.content.localPosition);

            }

        }
        protected void RestoreScrollBarPosition(string atomName)
        {
 //           Canvas.ForceUpdateCanvases();
            if (atomName != "")
            {
                if (_atomACEScrollBarRectPosition.ContainsKey(atomName)) scrollListSR.content.localPosition = _atomACEScrollBarRectPosition[atomName];
                else
                {
                    scrollListSR.verticalNormalizedPosition = 1f;
                    _atomACEScrollBarRectPosition.Add(atomName, scrollListSR.content.localPosition);
                }
            }
            else scrollListSR.verticalNormalizedPosition = 1f;
        }

        
        public static Color undressOn = new Color(210f / 255f, 245f / 255f, 12f / 255f);
        public static Color undressOff = new Color(245f / 255f, 213f / 255f, 34f / 255f);
        public static Color removeOn = new Color(144f / 255f, 237f / 255f, 240f / 255f);
        public static Color removeOff = new Color(144f / 255f, 240f / 255f, 182f / 255f);
        public static Color grey = new Color(0.8f, 0.8f, 0.8f);

        protected virtual void RefreshACEGroup(ActiveClothingLists currentACL)
        {
            if (currentACL != null)
            {
                if (_removeSelectedGroupState)
                {
                    _aceSelectedItemsGroup.aceItemRemoveButton.label = "Remove";
                    _aceSelectedItemsGroup.aceItemRemoveButton.buttonColor = removeOn;
                }
                else
                {
                    _aceSelectedItemsGroup.aceItemRemoveButton.label = "Wear";
                    _aceSelectedItemsGroup.aceItemRemoveButton.buttonColor = removeOff;
                }

                if (_undressSelectedGroupState)
                {
                    _aceSelectedItemsGroup.aceItemUndressButton.label = "Undress";
                    _aceSelectedItemsGroup.aceItemUndressButton.buttonColor = undressOff;
                }
                else
                {
                    _aceSelectedItemsGroup.aceItemUndressButton.label = "Dress";
                    _aceSelectedItemsGroup.aceItemUndressButton.buttonColor = undressOn;
                }

                if (currentACL.GetSelectedCount() > 0 && !_aceSelectedItemsGroup._aceItemSelectToggle.toggle.isOn)
                {
                    _suppressToggleGroupSelectOn = true;
                    _aceSelectedItemsGroup._aceItemSelectToggle.toggle.isOn = true;
                }
                else if (currentACL.GetSelectedCount() == 0)
                {
                    _aceSelectedItemsGroup._aceItemSelectToggle.toggle.isOn = false;
                    _aceSelectedItemsGroup.aceItemUndressButton.buttonColor = grey;
                    _aceSelectedItemsGroup.aceItemRemoveButton.buttonColor = grey;

                }
            }
            else
            {
                SetClothingItemCount(0);
                _suppressToggleGroupSelectOn = true;
                _aceSelectedItemsGroup._aceItemSelectToggle.toggle.isOn = false;
                _aceSelectedItemsGroup.aceItemRemoveButton.label = "Remove";
                _aceSelectedItemsGroup.aceItemRemoveButton.buttonColor = grey;
                _aceSelectedItemsGroup.aceItemUndressButton.label = "Undress";
                _aceSelectedItemsGroup.aceItemUndressButton.buttonColor = grey;
            }
        }

        protected void RefreshACEHeader()
        {
            _targetAtomTextField.text = _currentACEAtomName;
        }

        protected virtual void RefreshACE()
        {
            if (_atomACEClothingLists.ContainsKey(_currentACEAtomName))
            {
                SetClothingItemCount(_atomACEClothingLists[_currentACEAtomName]._activeClothingDCIs.Count);
                ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
                int index = 0;
                foreach (ActiveClothingEditorItemDisplay uidACEitem in _aceScrollListItems)
                {
                    uidACEitem.RefreshACEItem(currentACL, index);

                    index++;
                }
                RefreshACEGroup(currentACL);               
            }
            else
            {
                RefreshACEGroup(null);
            }
            RefreshACEHeader();
        }
        public void CreateUI()
        {
            float scrollList1Height = 200f;
            float scrollList1Width = 300f;
            float listXoffset = -(scrollList1Width / 4f);
            float selectGroupHeight = 20f;
            float refreshButtonWidth = 95f;
            float refreshButtonHeight = 30f;
            float atomNameTextFieldWidth = 200f;
            float atomNameTextFieldHeight = 30f;

            aceScrollListGO = _uiButtonCanvas.CreateScrollableButtonList(0f, (scrollList1Height / 2f) + _verticalStartOffset + selectGroupHeight, scrollList1Width, scrollList1Height, new Color(0.5f, 0.5f, 0.5f, 0.5f));
            _aceSelectedItemsGroup = new ActiveClothingEditorItemDisplay(_uiButtonCanvas, null, new Color(0.9f, 0.9f, 0.9f));

            _aceSelectedItemsGroup.aceItemContainerImageGO.transform.localPosition = new Vector3(0f, _verticalStartOffset + (selectGroupHeight / 2f), 0f);
            _aceSelectedItemsGroup._aceItemSelectToggle.label = "SELECTED CLOTHING ITEMS";

            _aceSelectedItemsGroup.aceItemRemoveButton.button.onClick.AddListener(delegate
            {
                try { RemoveSelectedButtonActivated(); }
                catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
            });

            _aceSelectedItemsGroup.aceItemUndressButton.button.onClick.AddListener(delegate
            {
                try { UndressSelectedButtonActivated(); }
                catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
            });
            _aceSelectedItemsGroup._aceItemSelectToggle.toggle.onValueChanged.AddListener((bool value) =>
            {
                try { SelectGroupToggleActivated(value); }
                catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
            });

            _refreshButton = _uiButtonCanvas.CreateUIButtonInCanvas("Clear Removed Clothing", (scrollList1Width / 2f) - (refreshButtonWidth / 2f), scrollList1Height + _verticalStartOffset + selectGroupHeight + (refreshButtonHeight / 2f), 0f, refreshButtonWidth, refreshButtonHeight, 12);
            _refreshButton.button.onClick.AddListener(delegate
            {
                try { RefreshClothing(_currentACEAtomName); }
                catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
            });

            _targetAtomTextField = _uiButtonCanvas.CreateTextFieldInCanvas(-(scrollList1Width / 2f) + (atomNameTextFieldWidth / 2f), scrollList1Height + _verticalStartOffset + selectGroupHeight, 0f, atomNameTextFieldWidth, atomNameTextFieldHeight);


            _targetAtomTextField.UItext.alignment = TextAnchor.MiddleLeft;
            _targetAtomTextField.UItext.fontSize = 14;

            _currentACEAtomName = "";
        }


        protected virtual void AddACEID(int listIndex)
        {
            ActiveClothingEditorItemDisplay aceItem = new ActiveClothingEditorItemDisplay(_uiButtonCanvas, aceScrollListGO, new Color(0f, 0f, 0f, 0f));
            aceItem.aceItemRemoveButton.button.onClick.AddListener(delegate
            {
                try { RemoveButtonActivated(listIndex); }
                catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
            });
            aceItem._aceItemSelectToggle.toggle.onValueChanged.AddListener((bool value) =>
            {
                try { SelectToggleActivated(listIndex, value); }
                catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
            });
            aceItem.aceItemUndressButton.button.onClick.AddListener(delegate
            {
                try { UndressButtonActivated(listIndex); }
                catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
            });
            _aceScrollListItems.Add(aceItem);
        }

        protected void SetClothingItemCount(int count)
        {
            if (_aceScrollListItems.Count < count)
            {
                int addItemCount = count - _aceScrollListItems.Count;
                int initialAceItemCount = _aceScrollListItems.Count;
                for (int i = 0; i < addItemCount; i++)
                {
                    int listIndex = initialAceItemCount + i;
                    AddACEID(listIndex);

                }
            }
            else if (_aceScrollListItems.Count > count)
            {
                int removeItemCount = _aceScrollListItems.Count - count;
                for (int i = 0; i < removeItemCount; i++)
                {
                    ActiveClothingEditorItemDisplay cei = _aceScrollListItems[_aceScrollListItems.Count - 1];

                    _aceScrollListItems.RemoveAt(_aceScrollListItems.Count - 1);
                    cei.DestroyClothingItem();
                }
            }
        }

        protected void SelectToggled(int index)
        {
            ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
            DAZClothingItem clothingItem = currentACL._activeClothingDCIs[index];

            currentACL.ToggleSelectedDCIState(clothingItem);

            SetUndressSelectedButtonState();
            SetRemoveSelectedButtonState();
        }

        private void SelectToggleActivated(int index, bool value)
        {
            ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
            DAZClothingItem clothingItem = currentACL._activeClothingDCIs[index];

            currentACL.UpdateSelectedDCIState(clothingItem, value);

            SetUndressSelectedButtonState();
            SetRemoveSelectedButtonState();
        }
        protected void SetRemoveSelectedButtonState()
        {
            bool val = false;
            ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
            Dictionary<DAZClothingItem, bool> selectedDCIs = currentACL._activeClothingSelected;

            bool noneSelected = true;

            foreach (DAZClothingItem dci in currentACL._activeClothingDCIs)
            {
                if (selectedDCIs.ContainsKey(dci) && selectedDCIs[dci])
                {
                    noneSelected = false;
                    if (dci.isActiveAndEnabled)
                    {
                        val = true;
                        break;
                    }

                }
            }
            _removeSelectedGroupState = (noneSelected || val);
        }
        protected void SetUndressSelectedButtonState()
        {
            bool val = false;
            ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
            Dictionary<DAZClothingItem, bool> selectedDCIs = currentACL._activeClothingSelected;

            bool noneSelected = true;

            foreach (DAZClothingItem dci in currentACL._activeClothingDCIs)
            {
                if (selectedDCIs.ContainsKey(dci) && selectedDCIs[dci] && currentACL._activeClothingCSC.ContainsKey(dci) && currentACL._activeClothingCSC[dci].simEnabledJSON.val && dci.isActiveAndEnabled)
                {
                    noneSelected = false;
                    if (!currentACL._activeClothingCSC[dci].clothSettings.BreakEnabled)
                    {
                        val = true;
                        break;
                    }

                }
            }
            _undressSelectedGroupState = (noneSelected || val);
        }
        protected void RemoveSelectedButtonActivated()
        {
            ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
            Dictionary<DAZClothingItem, bool> selectedDCIs = currentACL._activeClothingSelected;
            Atom atom = SuperController.singleton.GetAtomByUid(_currentACEAtomName);
            if (atom != null)
            {
                JSONStorable receiver = atom.GetStorableByID("geometry");
                if (receiver != null)
                {
                    bool targetRemoveClothingState = !_removeSelectedGroupState;
                    foreach (DAZClothingItem dci in currentACL._activeClothingDCIs)
                    {
                        if (selectedDCIs.ContainsKey(dci) && selectedDCIs[dci])
                        {
                            JSONStorableBool active = receiver.GetBoolJSONParam("clothing:" + dci.uid);
                            active.val = targetRemoveClothingState;
                        }
                    }
                }
            }
            SetRemoveSelectedButtonState();
            SetUndressSelectedButtonState();
        }

        protected void UndressSelectedButtonActivated()
        {
            ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
            Dictionary<DAZClothingItem, bool> selectedDCIs = currentACL._activeClothingSelected;

            Atom atom = SuperController.singleton.GetAtomByUid(_currentACEAtomName);
            if (atom != null)
            {
                bool targetUndressClothingState = !_undressSelectedGroupState;
                foreach (DAZClothingItem dci in currentACL._activeClothingDCIs)
                {

                    if (selectedDCIs.ContainsKey(dci) && selectedDCIs[dci] && dci.isActiveAndEnabled && _atomACEClothingLists[_currentACEAtomName]._activeClothingCSC.ContainsKey(dci))
                    {
                        ClothSimControl csc = _atomACEClothingLists[_currentACEAtomName]._activeClothingCSC[dci];
                        JSONStorable receiver = atom.GetStorableByID(csc.storeId);
                        if (receiver != null && csc.simEnabledJSON.val)
                        {
                            JSONStorableBool detached = receiver.GetBoolJSONParam("allowDetach");
                            if (detached != null)
                            {
                                if (!targetUndressClothingState) detached.val = true;
                                else
                                {
                                    if (detached.val) csc.Reset();
                                    detached.val = false;
                                }
                            }

                        }
                    }
                }
            }
            SetUndressSelectedButtonState();
        }

        protected void SelectGroupToggleActivated(bool activated)
        {
            if (_suppressToggleGroupSelectOn) _suppressToggleGroupSelectOn = false;
            else
            {
                ActiveClothingLists currentACL = _atomACEClothingLists[_currentACEAtomName];
                foreach (DAZClothingItem dci in currentACL._activeClothingDCIs)
                {
                    currentACL.UpdateSelectedDCIState(dci, activated);
                }

                SetUndressSelectedButtonState();
                SetRemoveSelectedButtonState();
            }

        }
        public void RefreshClothing(string atomName)
        {
            _atomACEClothingLists.Remove(atomName);
        }
        protected void RemoveButtonActivated(int index)
        {
            DAZClothingItem dci = _atomACEClothingLists[_currentACEAtomName]._activeClothingDCIs[index];
            Atom atom = SuperController.singleton.GetAtomByUid(_currentACEAtomName);
            if (atom != null)
            {
                JSONStorable receiver = atom.GetStorableByID("geometry");
                if (receiver != null)
                {
                    JSONStorableBool active = receiver.GetBoolJSONParam("clothing:" + dci.uid);
                    if (active != null) active.val = !active.val;
                }
            }
            SetRemoveSelectedButtonState();
            SetUndressSelectedButtonState();
        }
        protected void UndressButtonActivated(int index, bool resetSim = true)
        {
            DAZClothingItem dci = _atomACEClothingLists[_currentACEAtomName]._activeClothingDCIs[index];
            ClothSimControl csc = _atomACEClothingLists[_currentACEAtomName]._activeClothingCSC[dci];

            Atom atom = SuperController.singleton.GetAtomByUid(_currentACEAtomName);
            if (atom != null)
            {
                JSONStorable receiver = atom.GetStorableByID(csc.storeId);
                if (receiver != null)
                {
                    JSONStorableBool detached = receiver.GetBoolJSONParam("allowDetach");
                    if (detached != null)
                    {
                        if (detached.val && resetSim) csc.Reset();
                        detached.val = !detached.val;
                    }
                }
            }
            SetUndressSelectedButtonState();
        }

        

        public virtual void DestroyUI()
        {
            if (aceScrollListGO != null)
            {

                foreach (ActiveClothingEditorItemDisplay cei in _aceScrollListItems) cei.DestroyClothingItem();
                _aceScrollListItems.Clear();
                GameObject.Destroy(aceScrollListGO);
                aceScrollListGO = null;
            }
            if (_aceSelectedItemsGroup != null) _aceSelectedItemsGroup.DestroyClothingItem();
            if (_refreshButton != null) GameObject.Destroy(_refreshButton.gameObject);
            if (_targetAtomTextField != null) GameObject.Destroy(_targetAtomTextField.gameObject);
        }
    }

    public class ACEIDPlus : ActiveClothingEditorItemDisplay
    {
        public UIDynamicButton aceItemImageButton;
        public UIDynamicButton aceItemNameButton;
        public UIDynamicButton aceItemPresetsButton;
        public UIDynamicButton aceItemResetSimButton;

        private Texture2D lastLoadedTexture;
        private bool isLoadingTexture = false;

        private DCIButtonPointerBehaviour dciItemImagePointerBehaviour;
        private DCIButtonPointerBehaviour dciItemNamePointerBehaviour;

        public DAZClothingItem currentDAZClothingItem;
        public ACEIDPlus(GridsCanvas uiButtonCanvas, GameObject aceScrollListGO, Color panelColor) :base(uiButtonCanvas, aceScrollListGO, panelColor)
        {
        }

        protected override void CreateUI(GameObject aceScrollListGO, Color panelColor)
        {
            float horizontalOffset = 5f;

            if (aceScrollListGO != null)
            {
                GameObject buttonListContentGO = aceScrollListGO.transform.Find("buttonListViewportGO/buttonListContentGO").gameObject;

                aceItemContainerImageGO = UITools.CreateUIContainerImageInParentGO(buttonListContentGO.transform, 380f, 50f, panelColor);
                horizontalOffset = 0f;
            }
            else aceItemContainerImageGO = UITools.CreateUIContainerImageInParentGO(_uiButtonCanvas._canvas.gameObject.transform, 400f, 50f, panelColor);

            var imageButtonGO = UITools.CreateUIButtonInParent(aceItemContainerImageGO.transform);
            UITools.SetSizeAndAnchors(aceItemContainerImageGO, imageButtonGO, 50f, 50f, "left", 0, 0);
            aceItemImageButton = imageButtonGO.transform.GetComponent<UIDynamicButton>();
            aceItemImageButton.label = "";
            dciItemImagePointerBehaviour = aceItemImageButton.button.gameObject.AddComponent<DCIButtonPointerBehaviour>();

            var aceItemNameButtonGO = UITools.CreateUIButtonInParent(aceItemContainerImageGO.transform);
            aceItemNameButton = aceItemNameButtonGO.transform.GetComponent<UIDynamicButton>();
            UITools.SetSizeAndAnchors(aceItemContainerImageGO, aceItemNameButtonGO, 220f, 50f, "left", horizontalOffset + 50f, 0f);
            aceItemNameButton.buttonText.resizeTextMinSize = 8;
            aceItemNameButton.buttonText.resizeTextMaxSize = 16;
            aceItemNameButton.buttonText.resizeTextForBestFit = true;
            aceItemNameButton.button.image.sprite = null;
            dciItemNamePointerBehaviour = aceItemNameButton.button.gameObject.AddComponent<DCIButtonPointerBehaviour>();

            var aceItemRemoveButtonGO = UITools.CreateUIButtonInParent(aceItemContainerImageGO.transform);
            UITools.SetSizeAndAnchors(aceItemContainerImageGO, aceItemRemoveButtonGO, 50f, 25f, "topright", -60f,0f);
            aceItemRemoveButton = aceItemRemoveButtonGO.transform.GetComponent<UIDynamicButton>();
            aceItemRemoveButton.buttonText.fontSize = 11;
            aceItemRemoveButton.label = "Remove";
            aceItemRemoveButton.button.image.sprite = null;

            var acePresetsButtonGO = UITools.CreateUIButtonInParent(aceItemContainerImageGO.transform);
            UITools.SetSizeAndAnchors(aceItemContainerImageGO, acePresetsButtonGO, 50f, 25f, "bottomright", -60f, 0f);
            aceItemPresetsButton = acePresetsButtonGO.transform.GetComponent<UIDynamicButton>();
            aceItemPresetsButton.buttonText.fontSize = 11;
            aceItemPresetsButton.label = "Presets...";
            aceItemPresetsButton.button.image.sprite = null;

            var aceItemUndressButtonGO = UITools.CreateUIButtonInParent(aceItemContainerImageGO.transform);
            UITools.SetSizeAndAnchors(aceItemContainerImageGO, aceItemUndressButtonGO, 50f, 25f, "topright", -10f);
            aceItemUndressButton = aceItemUndressButtonGO.transform.GetComponent<UIDynamicButton>();
            aceItemUndressButton.buttonText.fontSize = 11;
            aceItemUndressButton.label = "Undress";
            aceItemUndressButton.button.image.sprite = null;

            var aceItemResetSimButtonGO = UITools.CreateUIButtonInParent(aceItemContainerImageGO.transform);
            UITools.SetSizeAndAnchors(aceItemContainerImageGO, aceItemResetSimButtonGO, 50f, 25f, "bottomright", -10f);
            aceItemResetSimButton = aceItemResetSimButtonGO.transform.GetComponent<UIDynamicButton>();
            aceItemResetSimButton.buttonText.fontSize = 11;
            aceItemResetSimButton.label = "Reset\nSim";
            aceItemResetSimButton.buttonColor = Color.red;
            aceItemResetSimButton.textColor = Color.white;
            aceItemResetSimButton.button.image.sprite = null;
        }

        public override void RefreshACEItem(ActiveClothingLists currentACL, int index)
        {
            DAZClothingItem clothingItem = currentACL._activeClothingDCIs[index];

            if ((clothingItem != currentDAZClothingItem || lastLoadedTexture==null) & !isLoadingTexture)
            {
                isLoadingTexture = true;
                clothingItem.GetThumbnail(delegate (Texture2D tex)
                {
                    
 //                   if (lastLoadedTexture != null) GameObject.Destroy(lastLoadedTexture);
                    lastLoadedTexture = tex;
                    Sprite mySprite = Sprite.Create(tex, new Rect(0.0f, 0.0f, tex.width, tex.height), new Vector2(0.5f, 0.5f), 100.0f);
                  
                    aceItemImageButton.button.image.sprite = mySprite;
                    isLoadingTexture = false;
                });
                currentDAZClothingItem = clothingItem;
            }
            dciItemNamePointerBehaviour.Init(clothingItem);
            dciItemImagePointerBehaviour.Init(clothingItem);

            aceItemNameButton.label = currentACL._activeClothingDisplayNames[clothingItem];

            if (currentACL.GetSelectedDCIState(clothingItem))
            {
                aceItemNameButton.buttonColor = Color.cyan;
            }
            else aceItemNameButton.buttonColor = ActiveClothingEditorDisplay.grey;
            aceItemImageButton.buttonColor = Color.white;
            aceItemPresetsButton.buttonColor = ActiveClothingEditorDisplay.grey;
            aceItemResetSimButton.buttonColor = Color.red;
            aceItemResetSimButton.textColor = Color.white;

            if (clothingItem.isActiveAndEnabled)
            {
                aceItemRemoveButton.label = "Remove";
                aceItemRemoveButton.buttonColor = ActiveClothingEditorDisplay.removeOn;
                if (ACEPlusDisplay.clothingPresets[clothingItem].Count > 0) aceItemPresetsButton.gameObject.SetActive(true);
                else aceItemPresetsButton.gameObject.SetActive(false);
            }
            else
            {
                aceItemRemoveButton.label = "Wear";
                aceItemRemoveButton.buttonColor = ActiveClothingEditorDisplay.removeOff;
                aceItemPresetsButton.gameObject.SetActive(false);
            }

            if (currentACL._activeClothingCSC.ContainsKey(clothingItem) && currentACL._activeClothingCSC[clothingItem].simEnabledJSON.val && clothingItem.isActiveAndEnabled)
            {
                aceItemUndressButton.gameObject.SetActive(true);
                aceItemResetSimButton.gameObject.SetActive(true);
                if (currentACL._activeClothingCSC[clothingItem].clothSettings.BreakEnabled)
                {
                    aceItemUndressButton.label = "Stop Undress";
                    aceItemUndressButton.buttonColor = ActiveClothingEditorDisplay.undressOn;
                }

                else
                {
                    aceItemUndressButton.label = "Undress";
                    aceItemUndressButton.buttonColor = ActiveClothingEditorDisplay.undressOff;
                }
            }
            else
            {
                aceItemUndressButton.gameObject.SetActive(false);
                aceItemResetSimButton.gameObject.SetActive(false);
            }

        }

        public override void DestroyClothingItem()
        {

            base.DestroyClothingItem();
//            if (lastLoadedTexture!=null) GameObject.Destroy(lastLoadedTexture);
        }
    }

    public class ActiveClothingEditorItemDisplay
    {
        protected GridsCanvas _uiButtonCanvas;
        public GameObject aceItemContainerImageGO;

        public UIDynamicToggle _aceItemSelectToggle;
        public UIDynamicButton aceItemRemoveButton;
        public UIDynamicButton aceItemUndressButton;

        public ActiveClothingEditorItemDisplay(GridsCanvas uiButtonCanvas, GameObject aceScrollListGO, Color panelColor)
        {
            try
            {
                _uiButtonCanvas = uiButtonCanvas;

                CreateUI(aceScrollListGO, panelColor);
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        protected virtual void CreateUI(GameObject aceScrollListGO, Color panelColor)
        {
            float horizontalOffset = 5f;

            if (aceScrollListGO != null)
            {
                GameObject buttonListContentGO = aceScrollListGO.transform.Find("buttonListViewportGO/buttonListContentGO").gameObject;

                aceItemContainerImageGO = UITools.CreateUIContainerImageInParentGO(buttonListContentGO.transform, 280f, 25f, panelColor);
                horizontalOffset = 0f;
            }
            else aceItemContainerImageGO = UITools.CreateUIContainerImageInParentGO(_uiButtonCanvas._canvas.gameObject.transform, 300f, 25f, panelColor);


            var _aceItemToggleGO = UITools.CreateToggle(aceItemContainerImageGO.transform, 0f, 0f, 180f, 25f);
            UITools.SetAnchors(aceItemContainerImageGO, _aceItemToggleGO, "left", horizontalOffset);
            _aceItemToggleGO.SetActive(true);
            _aceItemSelectToggle = _aceItemToggleGO.transform.GetComponent<UIDynamicToggle>();
            _aceItemSelectToggle.labelText.resizeTextMinSize = 8;
            _aceItemSelectToggle.labelText.resizeTextMaxSize = 13;
            _aceItemSelectToggle.labelText.resizeTextForBestFit = true;


            var aceItemRemoveButtonGO = UITools.CreateUIButtonInParent(aceItemContainerImageGO.transform);
            UITools.SetSizeAndAnchors(aceItemContainerImageGO, aceItemRemoveButtonGO, 50f, 20f, "right", -50f - (horizontalOffset * 3));
            aceItemRemoveButton = aceItemRemoveButtonGO.transform.GetComponent<UIDynamicButton>();
            aceItemRemoveButton.buttonText.fontSize = 11;
            aceItemRemoveButton.label = "Remove";

            var aceItemUndressButtonGO = UITools.CreateUIButtonInParent(aceItemContainerImageGO.transform);
            UITools.SetSizeAndAnchors(aceItemContainerImageGO, aceItemUndressButtonGO, 50f, 20f, "right", -horizontalOffset * 3);
            aceItemUndressButton = aceItemUndressButtonGO.transform.GetComponent<UIDynamicButton>();
            aceItemUndressButton.buttonText.fontSize = 11;
            aceItemUndressButton.label = "Undress";
        }

        public virtual void RefreshACEItem(ActiveClothingLists currentACL, int index)
        {
            DAZClothingItem clothingItem = currentACL._activeClothingDCIs[index];
            _aceItemSelectToggle.label = currentACL._activeClothingDisplayNames[clothingItem];
            if (clothingItem.isActiveAndEnabled)
            {
                aceItemRemoveButton.label = "Remove";
                aceItemRemoveButton.buttonColor = ActiveClothingEditorDisplay.removeOn;
            }
            else
            {
                aceItemRemoveButton.label = "Wear";
                aceItemRemoveButton.buttonColor = ActiveClothingEditorDisplay.removeOff;
            }

            if (currentACL._activeClothingCSC.ContainsKey(clothingItem) && currentACL._activeClothingCSC[clothingItem].simEnabledJSON.val && clothingItem.isActiveAndEnabled)
            {
                aceItemUndressButton.gameObject.SetActive(true);
                if (currentACL._activeClothingCSC[clothingItem].clothSettings.BreakEnabled)
                {
                    aceItemUndressButton.label = "Dress";
                    aceItemUndressButton.buttonColor = ActiveClothingEditorDisplay.undressOn;
                }

                else
                {
                    aceItemUndressButton.label = "Undress";
                    aceItemUndressButton.buttonColor = ActiveClothingEditorDisplay.undressOff;
                }
            }
            else aceItemUndressButton.gameObject.SetActive(false);

            _aceItemSelectToggle.toggle.isOn = currentACL.GetSelectedDCIState(clothingItem);
            _aceItemSelectToggle.backgroundColor = ActiveClothingEditorDisplay.grey;
        }


        public virtual void DestroyClothingItem()
        {
            if (aceItemRemoveButton!=null) aceItemRemoveButton.button.onClick.RemoveAllListeners();
            if (aceItemUndressButton != null) aceItemUndressButton.button.onClick.RemoveAllListeners();
            if (_aceItemSelectToggle != null) _aceItemSelectToggle.toggle.onValueChanged.RemoveAllListeners();
            GameObject.Destroy(aceItemContainerImageGO);
        }

    }


    static class PatreonFeatures
    {
        public const bool patreonContentEnabled = true;

        public static float CreateQLBButtons(SingleGridDisplay qlbGridDisplay, Vector2 qlbSize)
        {
            qlbGridDisplay.CreateUIButtons(UIAStorables.quickLaunchButtonGrid, 20f);
            qlbGridDisplay.UpdateButtonTextures();
            return (qlbSize.y * UIAStorables.quickLaunchButtonGrid.gridRows) + 20f;
        }

        public static Vector2 GridSize(UIAButtonGrid buttonGrid)
        {
            return new Vector2(buttonGrid.gridColumns, buttonGrid.gridRows);
        }

        public static bool GetNextFileSelection(FileReference fileRef, out string filePath, string lastSelectedFile, uFileBrowser.FileBrowserCallback callback)
        {
            filePath = "";
            
            string fileType = "vap";
            string prefixRemoval = null;

            UIAButton parentButton = fileRef.parentButton;
            UIAButtonOperation parentButtonOp = fileRef.parentButtonOperation;
            int filRefCategory = FileReferenceTypes.GetFileRefCategory(fileRef.fileReferenceType);

            if (filRefCategory == FileReferenceCategory.presetCat) prefixRemoval = "Preset_";
            else if (fileRef.fileReferenceType == FileReferenceTypes.scenePluginPreset || fileRef.fileReferenceType == FileReferenceTypes.sessionPluginPreset) prefixRemoval = "Plugins_";

            if (filRefCategory == FileReferenceCategory.sceneCat || filRefCategory == FileReferenceCategory.legacyPresetCat) fileType = "json";
            else if (fileRef.fileReferenceType == FileReferenceTypes.uiap) fileType = "uiap";

            if (fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.singleFile) filePath = fileRef.filePathJSString.val;
            else if (parentButton.GetButtonState() == 2) filePath = lastSelectedFile;
            else
            {
                string normalizedLastSelectedFile = "";
                if (lastSelectedFile != "") normalizedLastSelectedFile = FileManagerSecure.NormalizePath(lastSelectedFile);
                string folderName = fileRef.filePathJSString.val;
                string defaultFolder = fileRef.GetDefaultFolder();
                if (folderName == "") folderName = defaultFolder;

                if (fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.chooseFromFolder)
                {
                    SuperController.singleton.activeUI = SuperController.ActiveUI.None;
                    SuperController.singleton.ShowMainHUDAuto();
                    List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory(folderName, false, false, true);
                    if (folderName != defaultFolder) shortcuts.AddRange(FileManagerSecure.GetShortCutsForDirectory(defaultFolder, false, false, true));

                    if (parentButtonOp.GetFileReferenceTypes().Contains(FileReferenceTypes.addonPackagesFolder))
                    {
                        FileReference shortcutFilterFileRef = parentButtonOp.fileReferenceDict[FileReferenceTypes.addonPackagesFolder];

                        List<ShortCut> filteredShortcuts = new List<ShortCut>();
                        string filterPath = SuperController.singleton.NormalizeDirectoryPath(shortcutFilterFileRef.filePathJSString.val);

                        foreach (ShortCut sc in shortcuts)
                        {
                            if (sc.path.StartsWith(filterPath)) filteredShortcuts.Add(sc);
                            else if (!sc.path.StartsWith("AddonPackages")) filteredShortcuts.Add(sc);
                        }
                        shortcuts = filteredShortcuts;
                    }

                    SuperController.singleton.GetMediaPathDialog(callback, fileType, folderName, false, true, false, prefixRemoval, false, shortcuts);

                    if (parentButtonOp.buttonOpTypeJSEnum.val == UIAButtonOpType.saveLegacyLook || parentButtonOp.buttonOpTypeJSEnum.val == UIAButtonOpType.saveLegacyPose || parentButtonOp.buttonOpTypeJSEnum.val == UIAButtonOpType.saveLegacyPreset || (parentButtonOp.buttonOpCategoryJSEnum.val == UIAButtonCategory.presets && parentButtonOp.savePresetJSB.val))
                    {
                        string extension = "json";
                        if (parentButtonOp.buttonOpCategoryJSEnum.val == UIAButtonCategory.presets && parentButtonOp.savePresetJSB.val) extension = "vap";
                            // Update the browser to be a Save browser
                        uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
                        browser.SetTextEntry(true);
                        browser.fileEntryField.text = String.Format("{0}.{1}", ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString(), extension);
                        browser.ActivateFileNameField();
                    }
                    return false;
                }
                else if (fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.chooseFromBA)
                {
                    
                    BAInterop.SelectResourceFilename(fileRef.fileReferenceType, delegate (string fileName)
                    {
                        callback(fileName);
                    }, BADesktopDisplayModeEnum.leftDocked,"Select "+fileRef.GetFileRefTypeDescription(),BAKeepOpenOptionEnum.availableInactive);
                    if (UIAGlobals.vrActive && GameControlUI.gameControlUIMode!=GameControlUIModes.vrVAMUI)
                    {
                        GridsControlDisplay.CloseGrids();
 //                       GridsControlDisplay.MainButtonClickDown();
 //                       GridsControlDisplay.MainButtonClickFinish();
                    }

                }
                else if (fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.sequentialFromBA)
                {
                    BAInterop.GetOrderedFilteredResourceFilenames(fileRef.baFilter, delegate (List<string> resourceFileNames)
                    {
                        List<string> normalizedResourceFileNames = new List<string>();
                        foreach (string resourceFileName in resourceFileNames) normalizedResourceFileNames.Add(FileManagerSecure.NormalizePath(resourceFileName));
                        if (resourceFileNames.Count>0)
                        {
                            string fileName = "";
                            if (!normalizedResourceFileNames.Contains(lastSelectedFile) || lastSelectedFile == normalizedResourceFileNames.Last()) fileName = normalizedResourceFileNames[0];
                            else fileName = normalizedResourceFileNames[normalizedResourceFileNames.IndexOf(lastSelectedFile) + 1];

                            callback(fileName);
                        }
                        
                    });
                    return false;
                }
                else if (fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.randomFromBA)
                {
                    BAInterop.GetFilteredResourceFilenames(fileRef.baFilter, delegate (HashSet<string> resourceFileNames)
                    {
                        List<string> resourceFileNamesList = new List<string>(resourceFileNames);
                        if (resourceFileNamesList.Count > 0)
                        {
                            
                            var fileName = FileManagerSecure.NormalizePath( resourceFileNamesList[UnityEngine.Random.Range(0, resourceFileNamesList.Count)]);
                            if (resourceFileNamesList.Count > 1 && lastSelectedFile != "")
                            {
                                while (fileName == lastSelectedFile)
                                {
                                    fileName = FileManagerSecure.NormalizePath(resourceFileNamesList[UnityEngine.Random.Range(0, resourceFileNamesList.Count)]);
                                }
                            }
                            
                            callback(fileName);
                        }
                        
                    });
                    return false;
                }
                else
                {
                    List<string> filesList;
                    
                    if (fileRef.includeSubFoldersJSB.val) filesList = FileUtils.GetFilesAtPathRecursive(folderName, "*." + fileType).ToList();
                    else filesList = new List<string>(FileManagerSecure.GetFiles(folderName, "*." + fileType));
                    for (int i = 0; i < filesList.Count; i++) filesList[i] = FileManagerSecure.NormalizePath(filesList[i]);
                    filesList.Sort();

                    if (fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.alphabeticalFromFolder && filesList.Count() > 0)
                    {
                        if (!filesList.Contains(lastSelectedFile) || lastSelectedFile == filesList.Last()) filePath = filesList[0];
                        else filePath = filesList[filesList.IndexOf(lastSelectedFile) + 1];
                    }
                    else if (fileRef.fileSelectionModeJSEnum.val == FileSelectionMode.randomFromFolder && filesList.Count() > 0)
                    {
                        filePath = filesList[UnityEngine.Random.Range(0, filesList.Count)];
                        if (filesList.Count > 1 && lastSelectedFile != "")
                        {
                            while (filePath == lastSelectedFile)
                            {
                                filePath = filesList[UnityEngine.Random.Range(0, filesList.Count)];
                            }
                        }
                    }
                }

            }
            return true;
        }

         
        public static void GridLayoutFeature(List<string> buttonGridLayoutChoices)
        {
            buttonGridLayoutChoices.Add("4x5"); buttonGridLayoutChoices.Add("5x4"); buttonGridLayoutChoices.Add("5x5");
            buttonGridLayoutChoices.Add("5x6"); buttonGridLayoutChoices.Add("6x5"); buttonGridLayoutChoices.Add("6x6");
            buttonGridLayoutChoices.Add("7x6"); buttonGridLayoutChoices.Add("6x7"); buttonGridLayoutChoices.Add("7x7");
            buttonGridLayoutChoices.Add("8x7"); buttonGridLayoutChoices.Add("7x8"); buttonGridLayoutChoices.Add("8x8");
            buttonGridLayoutChoices.Add("9x8"); buttonGridLayoutChoices.Add("8x9"); buttonGridLayoutChoices.Add("9x9");

        }
        private static int GetGeometryIndex(JSONArray storablesJSONArray)
        {
            int geometryIndex = -1;
            for (int index = 0; index < storablesJSONArray.Count; index++)
            {
                if ((string)storablesJSONArray[index]["id"] == "geometry") geometryIndex = index;
            }
            return (geometryIndex);
        }

        public static int GetMergeClothingButtonState(Atom atom, FileReference fileRef)
        {
            string presetFileName = fileRef.GetCurrentFile(atom.name);

            if (!FileManagerSecure.FileExists(presetFileName)) return ButtonState.inactive;

            if (!UIAButton.presetMergeClothingGeometryIDs.ContainsKey(presetFileName)) PopulateMergeClothingGeometryIDs(presetFileName, atom);

            List<string> mergeClothingGeometryIDs = UIAButton.presetMergeClothingGeometryIDs[presetFileName];

            JSONStorable atomGeometryStorable = atom.GetStorableByID("geometry");

            int clothingItemCount = 0;
            foreach (string clothingItem in mergeClothingGeometryIDs)
            {
                JSONStorableBool clothingItemStateJSON = atomGeometryStorable.GetBoolJSONParam(clothingItem);
                if (clothingItemStateJSON != null)
                {
                    clothingItemCount++;
                    if (!clothingItemStateJSON.val) return ButtonState.inactive;
                }
            }
            if (clothingItemCount > 0)
            {
                if (!UIAButton.cachedClothingPresetFiles.ContainsKey(presetFileName)) UIAButton.cachedClothingPresetFiles.Add(presetFileName, SuperController.singleton.LoadJSON(presetFileName).AsObject);
                JSONClass clothingPresetJSON = UIAButton.cachedClothingPresetFiles[presetFileName];
                foreach (JSONClass presetStorableJC in clothingPresetJSON["storables"].AsArray)
                {
                    if (presetStorableJC["id"] != "geometry")
                    {
                        JSONStorable atomStorable = atom.GetStorableByID(presetStorableJC["id"]);
                        if (atomStorable != null && !ClothingStorableMatchesPreset(presetStorableJC, atomStorable))  return ButtonState.inactive;
                    }
                }
                return ButtonState.active;

            }
            return ButtonState.inactive;
        }
        private static bool ClothingStorableMatchesPreset(JSONClass presetStorableJC, JSONStorable atomClothingStorable)
        {
            
            foreach (var boolParamName in atomClothingStorable.GetBoolParamNames())
            {
                if (presetStorableJC[boolParamName] != null)
                {
                    if (presetStorableJC[boolParamName].AsBool != atomClothingStorable.GetBoolParamValue(boolParamName))
                    {                        
                        return false;
                    }
                }
            }

            foreach (var urlParamName in atomClothingStorable.GetUrlParamNames())
            {
                if (presetStorableJC[urlParamName] != null)
                {
                    string presetURL = presetStorableJC[urlParamName];
                    string storableURL = atomClothingStorable.GetUrlParamValue(urlParamName);
                    if (presetURL != storableURL)
                    {
                        if (!(string.IsNullOrEmpty(presetURL) && string.IsNullOrEmpty(storableURL)))
                        {
                            return false;
                        }
                    }
                }
            }

            foreach (var colorParamName in atomClothingStorable.GetColorParamNames())
            {
                if (presetStorableJC[colorParamName] != null)
                {
                    if (presetStorableJC[colorParamName]["h"].AsFloat != atomClothingStorable.GetColorParamValue(colorParamName).H) return false;
                    if (presetStorableJC[colorParamName]["s"].AsFloat != atomClothingStorable.GetColorParamValue(colorParamName).S) return false;
                    if (presetStorableJC[colorParamName]["v"].AsFloat != atomClothingStorable.GetColorParamValue(colorParamName).V) return false;
                }
            }

            foreach (var floatParamName in atomClothingStorable.GetFloatParamNames())
            {
                if (presetStorableJC[floatParamName] != null)
                {
                    if (presetStorableJC[floatParamName].AsFloat != atomClothingStorable.GetFloatParamValue(floatParamName))
                    {
                        return false;
                    }
                }
            }

            foreach (var stringParamName in atomClothingStorable.GetStringParamNames())
            {
                if (presetStorableJC[stringParamName] != null)
                {
                    string presetString = presetStorableJC[stringParamName];
                    string storableString = atomClothingStorable.GetStringParamValue(stringParamName);

                    if (presetString != storableString)
                    {
                        if (!(string.IsNullOrEmpty(presetString) && string.IsNullOrEmpty(storableString)))
                        {
                            return false;
                        }
                    }
                }
            }


            return true;
        }
        public static int GetUndressClothingButtonState(Atom atom, FileReference fileRef)
        {
            string presetFileName = fileRef.GetCurrentFile(atom.name);

            if (!FileManagerSecure.FileExists(presetFileName)) return ButtonState.inactive;

            if (!UIAButton.presetMergeClothingGeometryIDs.ContainsKey(presetFileName)) PopulateMergeClothingGeometryIDs(presetFileName, atom);

            List<string> mergeClothingGeometryIDs = UIAButton.presetMergeClothingGeometryIDs[presetFileName];
            Dictionary<string, string> undressClothingGeometryIDsToSimStorables = UIAButton.presetUndressClothingGeometryIDsToSimStorables[presetFileName];

            JSONStorable atomGeometryStorable = atom.GetStorableByID("geometry");

            int clothingItemCount = 0;
            foreach (string clothingItem in mergeClothingGeometryIDs)
            {

                JSONStorableBool clothingItemStateJSON = atomGeometryStorable.GetBoolJSONParam(clothingItem);
                //                
                if (clothingItemStateJSON != null && clothingItemStateJSON.val)
                {
                    JSONStorable simClothingStorable = atom.GetStorableByID(undressClothingGeometryIDsToSimStorables[clothingItem]);
                    if (simClothingStorable != null)
                    {
                        JSONStorableBool allowDetachJSB = simClothingStorable.GetBoolJSONParam("allowDetach");
                        clothingItemCount++;
                        if (!allowDetachJSB.val) return ButtonState.inactive;
                    }
                }
            }
            if (clothingItemCount > 0) return ButtonState.active;
            return ButtonState.inactive;
        }

        public static void ClothingPresetActions(Atom atom, int mode, string presetFile, bool reverse)
        {
            JSONNode presetJSON = UIAGlobals.mvrScript.LoadJSON(presetFile);
            string presetPackageName = "";
            if (presetFile.Contains(":")) presetPackageName = presetFile.Substring(0, presetFile.IndexOf(':'));

            PopulateMergeClothingGeometryIDs(presetFile, atom);
            List<string> mergeClothingGeometryIDs = UIAButton.presetMergeClothingGeometryIDs[presetFile];

            string presetJSONString = presetJSON.ToString();
            if (presetJSONString.Contains("SELF:"))
            {
                presetJSONString = presetJSONString.Replace("SELF:", presetPackageName + ":");
                presetJSON = (JSONClass)JSON.Parse(presetJSONString);
            }

            JSONArray presetStorablesJSON = presetJSON["storables"]?.AsArray;
            string presetStorablesString = presetStorablesJSON.ToString();

            if (presetStorablesJSON == null)
            {
                SuperController.LogMessage("UIAssist: Preset File missing 'storables' array");
                return;
            }

            int geometryIndex = GetGeometryIndex(presetStorablesJSON);
            if (geometryIndex > -1 && (mode == ClothingActionMode.merge || mode == ClothingActionMode.remove))
            {
                JSONStorable atomGeometryStorable = atom.GetStorableByID("geometry");
                JSONArray clothingJSON = presetStorablesJSON[geometryIndex]["clothing"].AsArray;
                if (atomGeometryStorable != null && clothingJSON != null)
                {
                    for (int clothingIndex = 0; clothingIndex < clothingJSON.Count; clothingIndex++)
                    {
                        if (clothingJSON[clothingIndex]["enabled"]==null || clothingJSON[clothingIndex]["enabled"].AsBool)
                        {
                            string presetClothingItemID = clothingJSON[clothingIndex]["id"];
                            string clothingGeometryID = GetClothingGeometryID(presetClothingItemID, atomGeometryStorable);
                            if (clothingGeometryID == null)
                            {
                                SuperController.LogMessage("UIAssist: Preset contains unavailable clothing item");
                                return;
                            }
                            else
                            {
                                if (mode == ClothingActionMode.merge) mergeClothingGeometryIDs.Add(clothingGeometryID);
                                JSONStorableBool clothingEnabled = atomGeometryStorable.GetBoolJSONParam(clothingGeometryID);
                                if ((string)clothingJSON[clothingIndex]["enabled"] == "true" && mode == ClothingActionMode.remove) clothingEnabled.val = false;
                                else if (clothingEnabled != null && mode == ClothingActionMode.merge) clothingEnabled.val = !reverse;
                            }
                            if (presetClothingItemID.Contains(":"))
                            {
                                string presetClothingItemPackageName = presetClothingItemID.Substring(0, presetClothingItemID.IndexOf(':'));
                                if (presetClothingItemPackageName.EndsWith(".latest") && presetStorablesString.Contains(presetClothingItemPackageName))
                                {
                                    string temp = clothingGeometryID.Substring(9);
                                    string temp2 = temp.Substring(0, temp.IndexOf(':'));
                                    presetStorablesString = presetStorablesString.Replace(presetClothingItemPackageName, temp2);

                                }
                            }
                        }
                        
                    }
                }
            }
            presetStorablesJSON = (JSONArray)JSON.Parse(presetStorablesString);

            UIAGlobals.mvrScript.StartCoroutine(MergeClothingStorables(atom, presetStorablesJSON, mode, reverse));
        }

        private static IEnumerator MergeClothingStorables(Atom atom, JSONArray storablesJSONArray, int mode, bool reverse)
        {
            float maxWait = 5f;
            float waitedFor = 0f;
            for (int index = 0; index < storablesJSONArray.Count; index++)
            {
                if ((string)storablesJSONArray[index]["id"] != "geometry" && mode != ClothingActionMode.remove)
                {
                    JSONStorable receiver = atom.GetStorableByID((string)storablesJSONArray[index]["id"]);
                    while (receiver == null && waitedFor < maxWait)
                    {
                        yield return new WaitForSeconds(0.01f);
                        waitedFor += 0.01f;
                        receiver = atom.GetStorableByID((string)storablesJSONArray[index]["id"]);
                    }
                    try
                    {
                        if (receiver != null)
                        {
                            foreach (string boolParamName in receiver.GetBoolParamNames())
                            {
                                if (storablesJSONArray[index][boolParamName] != null && mode == ClothingActionMode.merge) { receiver.GetBoolJSONParam(boolParamName).val = ((string)storablesJSONArray[index][boolParamName] == "true"); }
                                if (boolParamName == "allowDetach" && mode == ClothingActionMode.undress) { receiver.GetBoolJSONParam(boolParamName).val = !reverse; }
                                if (boolParamName == "simEnabled" && mode == ClothingActionMode.resetSim) { receiver.CallAction("Reset"); }
                            }
                            if (mode == ClothingActionMode.merge && !reverse)
                            {
                                foreach (string colorParamName in receiver.GetColorParamNames())
                                {
                                    if (storablesJSONArray[index][colorParamName] != null)
                                    {
                                        HSVColor color;
                                        color.H = storablesJSONArray[index][colorParamName]["h"].AsFloat;
                                        color.S = storablesJSONArray[index][colorParamName]["s"].AsFloat;
                                        color.V = storablesJSONArray[index][colorParamName]["v"].AsFloat;
                                        receiver.GetColorJSONParam(colorParamName).val = color;
                                    }
                                }
                                foreach (string floatParamName in receiver.GetFloatParamNames()) { if (storablesJSONArray[index][floatParamName] != null) { receiver.GetFloatJSONParam(floatParamName).val = storablesJSONArray[index][floatParamName].AsFloat; } }
                                foreach (string stringParamName in receiver.GetStringParamNames()) { if (storablesJSONArray[index][stringParamName] != null) { receiver.GetStringJSONParam(stringParamName).val = storablesJSONArray[index][stringParamName]; } }
                                foreach (string stringChooserParamName in receiver.GetStringChooserParamNames()) { if (storablesJSONArray[index][stringChooserParamName] != null) { receiver.GetStringChooserJSONParam(stringChooserParamName).val = storablesJSONArray[index][stringChooserParamName]; } }
                                foreach (string urlParamName in receiver.GetUrlParamNames())
                                {
                                    if (storablesJSONArray[index][urlParamName] != null) receiver.GetUrlJSONParam(urlParamName).val = storablesJSONArray[index][urlParamName];
                                }
                                foreach (string vector3ParamName in receiver.GetVector3ParamNames())
                                {
                                    if (storablesJSONArray[index][vector3ParamName] != null) { Vector3 vector; JSONArray vector3Array = storablesJSONArray[index][vector3ParamName].AsArray; vector.x = vector3Array[0].AsFloat; vector.y = vector3Array[1].AsFloat; vector.z = vector3Array[2].AsFloat; receiver.GetVector3JSONParam(vector3ParamName).val = vector; }
                                }
                            }
                        }
                    }
                    catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                }
            }
        }

        private static string GetClothingGeometryID(string presetClothingID, JSONStorable atomGeometryStorable)
        {
            string clothingID = null;

            if (presetClothingID.EndsWith(".vam") || atomGeometryStorable.GetBoolParamNames().Contains("clothing:" + presetClothingID))
            {
                if (presetClothingID.Contains(":"))
                {
                    // VAR packaged clothing items
                    int index = presetClothingID.IndexOf(':');
                    string presetPackageName = presetClothingID.Substring(0, index);
                    string presetFileName = presetClothingID.Substring(index + 1);
                    if (presetPackageName.EndsWith(".latest"))
                    {
                        string presetPackageNameWithoutVersion = presetPackageName.Substring(0, presetPackageName.Length - 6);
                        int highestClothingVARVersion = 0;
                        foreach (string atomGeometryBoolID in atomGeometryStorable.GetBoolParamNames())
                        {
                            if (atomGeometryBoolID.StartsWith("clothing:" + presetPackageNameWithoutVersion))
                            {
                                string temp = atomGeometryBoolID.Substring(9 + presetPackageNameWithoutVersion.Length);
                                string versionString = temp.Substring(0, temp.IndexOf(':'));
                                int clothingGeometryVersion = int.Parse(versionString);
                                if (clothingGeometryVersion > highestClothingVARVersion) highestClothingVARVersion = clothingGeometryVersion;
                            }
                        }
                        if (highestClothingVARVersion > 0) presetClothingID = presetPackageNameWithoutVersion + highestClothingVARVersion.ToString() + ":" + presetFileName;
                        else SuperController.LogError("UIAssist: GetClothingGeometryID did not find a clothing item in Geometry that matched the preset ID " + presetClothingID);
                    }

                }
                clothingID = "clothing:" + presetClothingID;
            }
            else
            {
                // Handles some non VAR packaged legacy clothing items (this may be redundant now?)
                string creatorName = "";
                if (presetClothingID.Contains(":"))
                {
                    int index = presetClothingID.IndexOf(':');
                    creatorName = presetClothingID.Substring(0, index);
                    presetClothingID = presetClothingID.Substring(index + 1);
                }
                foreach (string atomGeometryBoolID in atomGeometryStorable.GetBoolParamNames())
                {
                    if (atomGeometryBoolID.StartsWith("clothing:") && atomGeometryBoolID.Contains(presetClothingID) && (creatorName == "" || atomGeometryBoolID.StartsWith("clothing:" + creatorName)))
                    {
                        clothingID = atomGeometryBoolID;
                        break;
                    }
                }
            }
            return (clothingID);
        }
        private static void PopulateMergeClothingGeometryIDs(string presetFile, Atom atom)
        {
            if (!FileManagerSecure.FileExists(presetFile))
            {
                SuperController.LogMessage("UIAssist: Clothing preset '" + presetFile + "' does not exist or cant be found.");
                return;
            }
            JSONNode presetJSON = UIAGlobals.mvrScript.LoadJSON(presetFile);
            JSONArray storablesJSON = presetJSON["storables"]?.AsArray;

            List<string> mergeClothingGeometryIDs = new List<string>();
            Dictionary<string, string> undressClothGeometryIDToSimStorable = new Dictionary<string, string>();

            if (UIAButton.presetMergeClothingGeometryIDs.ContainsKey(presetFile)) UIAButton.presetMergeClothingGeometryIDs.Remove(presetFile);
            if (UIAButton.presetUndressClothingGeometryIDsToSimStorables.ContainsKey(presetFile)) UIAButton.presetUndressClothingGeometryIDsToSimStorables.Remove(presetFile);


            if (storablesJSON == null)
            {
                SuperController.LogMessage("UIAssist: Preset File missing 'storables' array");
                return;
            }
            for (int index = 0; index < storablesJSON.Count; index++)
            {
                if ((string)storablesJSON[index]["id"] == "geometry")
                {
                    JSONStorable atomGeometryStorable = atom.GetStorableByID("geometry");

                    JSONArray clothingJSON = storablesJSON[index]["clothing"].AsArray;

                    if (atomGeometryStorable != null && clothingJSON != null)
                    {
                        for (int clothingIndex = 0; clothingIndex < clothingJSON.Count; clothingIndex++)
                        {
                            if (clothingJSON[clothingIndex]["enabled"] == null || clothingJSON[clothingIndex]["enabled"].AsBool)
                            {
                                string clothingGeometryID = GetClothingGeometryID(clothingJSON[clothingIndex]["id"], atomGeometryStorable);
                                if (clothingGeometryID == null)
                                {
                                    SuperController.LogMessage("UIAssist: Preset contains unavailable clothing item");
                                    return;
                                }
                                else mergeClothingGeometryIDs.Add(clothingGeometryID);

                                if (clothingJSON[clothingIndex]["internalId"] != null) undressClothGeometryIDToSimStorable.Add(clothingGeometryID, clothingJSON[clothingIndex]["internalId"] + "Sim");
                                else
                                {
                                    string simStorableID = clothingGeometryID;
                                    if (simStorableID.EndsWith(".vam")) simStorableID = simStorableID.Substring(0, simStorableID.Length - 4);
                                    if (simStorableID.Contains("/")) simStorableID = simStorableID.Substring(simStorableID.LastIndexOf("/") + 1);

                                    undressClothGeometryIDToSimStorable.Add(clothingGeometryID, simStorableID + "Sim");
                                }
                            }
                            
                        }
                    }
                    break;
                }
            }


            UIAButton.presetMergeClothingGeometryIDs.Add(presetFile, mergeClothingGeometryIDs);
            UIAButton.presetUndressClothingGeometryIDsToSimStorables.Add(presetFile, undressClothGeometryIDToSimStorable);
        }

        public static void RemoveTagClothingItems(Atom atom, Dictionary<DAZClothingItem, List<string>> activeClothingTags, List<String> tagsToRemove)
        {
            JSONStorable receiver = atom.GetStorableByID("geometry");
            foreach (KeyValuePair<DAZClothingItem, List<string>> kvp in activeClothingTags)
            {
                List<string> tagIntersects = kvp.Value.Intersect(tagsToRemove).ToList();
                if (tagIntersects.Count > 0)
                {
                    JSONStorableBool boolTarget = receiver.GetBoolJSONParam("clothing:" + kvp.Key.uid);
                    if (boolTarget != null) boolTarget.val = false;
                }
            }
        }
        public static void AppPresetSuppressScaleLoad(Atom atom, FileReference fileRef, bool mergeLoad)
        {

            string presetPackageName = "";
            string presetFilePath = fileRef.currentActionSelectedFile;

            if (presetFilePath.Contains(":")) presetPackageName = presetFilePath.Substring(0, presetFilePath.IndexOf(':'));

            JSONClass presetJSON = (JSONClass)SuperController.singleton.LoadJSON(SuperController.singleton.NormalizePath(presetFilePath));
            JSONArray presetStorablesJSON = (JSONArray)presetJSON["storables"];

            JSONStorableFloat atomScale = atom.GetStorableByID("rescaleObject").GetFloatJSONParam("scale");
            foreach (JSONClass storableJSON in presetStorablesJSON)
            {
                if (storableJSON["id"].Value == "rescaleObject") storableJSON["scale"].AsFloat = atomScale.val;

            }
            if (presetPackageName != "")
            {
                string presetJSONString = presetJSON.ToString();
                if (presetJSONString.Contains("SELF:"))
                {
                    presetJSONString = presetJSONString.Replace("SELF:", presetPackageName + ":");
                    presetJSON = (JSONClass)JSON.Parse(presetJSONString);
                }
            }

            JSONStorable presetJS = atom.GetStorableByID("AppearancePresets");
            PresetManager pm = presetJS.GetComponentInChildren<PresetManager>();
            atom.SetLastRestoredData(presetJSON, true, true);
            pm.LoadPresetFromJSON(presetJSON, mergeLoad);
        }
        private static string ReplaceSelfReference(string textureUrl, string presetPackageName)
        {
            if (textureUrl.Contains("SELF:"))
            {
                textureUrl = textureUrl.Replace("SELF:", presetPackageName + ":");
            }
            return (textureUrl);
        }

        public static void SkinPresetDecalLoad(Atom atom, FileReference fileRef, SkinPresetDecalComponent skinPresetDecalComponent)
        {
            string presetPackageName = "";
            string presetFilePath = fileRef.currentActionSelectedFile;

            if (presetFilePath.Contains(":")) presetPackageName = presetFilePath.Substring(0, presetFilePath.IndexOf(':'));

            JSONClass skinPresetJSON = (JSONClass)SuperController.singleton.LoadJSON(SuperController.singleton.NormalizePath(presetFilePath));
            JSONClass skinPresetTextureJSON = new JSONClass();
            foreach (JSONClass storable in (JSONArray)skinPresetJSON["storables"])
            {
                if (storable["id"].Value == "textures")
                {
                    skinPresetTextureJSON = storable;
                    break;
                }
            }

            JSONStorable textureStorable = atom.GetStorableByID("textures");

            if (skinPresetDecalComponent.decalFaceLoadJSBool.val)
            {
                JSONStorableUrl texturePath = textureStorable.GetUrlJSONParam("faceDecalUrl");
                string newTexturePath = ReplaceSelfReference(skinPresetTextureJSON["faceDecalUrl"].Value, presetPackageName);
                if (newTexturePath.Contains(".latest:")) texturePath.val = FileUtils.GetLatestVARPath(newTexturePath);
                else texturePath.val = newTexturePath;

            }
            if (skinPresetDecalComponent.decalBodyLoadJSBool.val)
            {
                JSONStorableUrl texturePath = textureStorable.GetUrlJSONParam("torsoDecalUrl");
                string newTexturePath = ReplaceSelfReference(skinPresetTextureJSON["torsoDecalUrl"].Value, presetPackageName);
                if (newTexturePath.Contains(".latest:")) texturePath.val = FileUtils.GetLatestVARPath(newTexturePath);
                else texturePath.val = newTexturePath;
            }
            if (skinPresetDecalComponent.decalLimbsLoadJSBool.val)
            {
                JSONStorableUrl texturePath = textureStorable.GetUrlJSONParam("limbsDecalUrl");
                string newTexturePath = ReplaceSelfReference(skinPresetTextureJSON["limbsDecalUrl"].Value, presetPackageName);
                if (newTexturePath.Contains(".latest:")) texturePath.val = FileUtils.GetLatestVARPath(newTexturePath);
                else texturePath.val = newTexturePath;
            }
            if (skinPresetDecalComponent.decalGenLoadJSBool.val)
            {
                JSONStorableUrl texturePath = textureStorable.GetUrlJSONParam("genitalsDecalUrl");
                string newTexturePath = ReplaceSelfReference(skinPresetTextureJSON["genitalsDecalUrl"].Value, presetPackageName); ;
                if (newTexturePath.Contains(".latest:")) texturePath.val = FileUtils.GetLatestVARPath(newTexturePath);
                else texturePath.val = newTexturePath;
            }

        }

    }

}
